/*
 *       Copyright (C) 2002-12 Luca Deri <deri@ntop.org>
 *
 *                     http://www.ntop.org/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* ************************************************************************

   History:

   1.0    [06/02]        Initial release
   1.3    [07/02]        First public release

   ************************************************************************ */

#include "nprobe.h"

#define BLANK_SPACES               "                     "

/* #define HASH_DEBUG */

/* #define TIME_PROTECTION  */

#define MAX_SAMPLE_RATE    ((u_short)-1)

/* *************************************** */

/*
  #define OPTION_TEMPLATE "%SYSTEM_ID %SAMPLING_INTERVAL %SAMPLING_ALGORITHM %TOTAL_BYTES_EXP %TOTAL_PKTS_EXP %TOTAL_FLOWS_EXP %FLOW_ACTIVE_TIMEOUT %FLOW_INACTIVE_TIMEOUT"
*/

#define V9_OPTION_TEMPLATE "%TOTAL_FLOWS_EXP %TOTAL_PKTS_EXP"

/* IMPORTANT: when you modify it please also change exportBucketToNetflowV5 */
#define DEFAULT_V9_IPV4_TEMPLATE "%IPV4_SRC_ADDR %IPV4_DST_ADDR %IPV4_NEXT_HOP %INPUT_SNMP %OUTPUT_SNMP %IN_PKTS %IN_BYTES %FIRST_SWITCHED " \
  "%LAST_SWITCHED %L4_SRC_PORT %L4_DST_PORT %TCP_FLAGS %PROTOCOL %SRC_TOS %SRC_AS %DST_AS %IPV4_SRC_MASK %IPV4_DST_MASK"

#define DEFAULT_IPFIX_IPV4_TEMPLATE "%IPV4_SRC_ADDR %IPV4_DST_ADDR %IPV4_NEXT_HOP %INPUT_SNMP %OUTPUT_SNMP %IN_PKTS %IN_BYTES %FLOW_START_MILLISECONDS " \
  "%FLOW_END_MILLISECONDS %L4_SRC_PORT %L4_DST_PORT %TCP_FLAGS %PROTOCOL %SRC_TOS %SRC_AS %DST_AS %IPV4_SRC_MASK %IPV4_DST_MASK"

#define DEFAULT_V9_OPTION_TEMPLATE_ID (DEFAULT_TEMPLATE_ID+1)
#define TEMPLATE_PACKETS_DELTA        10

static void initDefaults(void);

/* *********** Globals ******************* */

#ifdef HAVE_PF_RING
#include "pro/pf_ring.c"
#endif

/* ****************************************************** */

/* Forward */
static void printStats(int force);
static int parseOptions(int argc, char* argv[], u_int8_t reparse_options);
static void compileTemplates(u_int8_t reloadTemplate);

static int argc_;
static char **argv_;

#ifdef HAVE_OPTRESET
extern int optreset; /* defined by BSD, but not others */
#endif

typedef void *(*pthread_start_routine)(void*);

static const struct option long_options[] = {
  { "all-collectors",                   required_argument,       NULL, 'a' },
  { "as-list",                          required_argument,       NULL, 'A' },
  { "verbose",                          required_argument,       NULL, 'b' },
  { "count-delay",                      required_argument,       NULL, 'B' },
  { "local-hosts-only",                 no_argument,             NULL, 'c' },
  { "flow-lock",                        required_argument,       NULL, 'C' },
  { "idle-timeout",                     required_argument,       NULL, 'd' },
  { "dump-format",                      required_argument,       NULL, 'D' },
  { "flow-delay",                       required_argument,       NULL, 'e' },
  { "netflow-engine",                   required_argument,       NULL, 'E' },
  { "bpf-filter",                       required_argument,       NULL, 'f' },
  { "dump-frequency",                   required_argument,       NULL, 'F' },
  { "pid-file",                         required_argument,       NULL, 'g' },
#ifndef WIN32
  { "daemon-mode",                      no_argument,             NULL, 'G' },
#endif
  { "help",                             no_argument,             NULL, 'h' },
  { "interface",                        required_argument,       NULL, 'i' },
  { "syslog",                           required_argument,       NULL, 'I' },
  { "queue-timeout",                    required_argument,       NULL, 'l' },
  { "local-networks",                   required_argument,       NULL, 'L' },
  { "min-num-flows",                    required_argument,       NULL, 'm' },
  { "max-num-flows",                    required_argument,       NULL, 'M' },
  { "collector",                        required_argument,       NULL, 'n' },
  { "rebuild-hash",                     no_argument,             NULL, 'N' }, /* Obsolete */
  { "flows-intra-templ",                required_argument,       NULL, 'o' },
  { "num-threads",                      required_argument,       NULL, 'O' },
  { "aggregation",                      required_argument,       NULL, 'p' },
  { "dump-path",                        required_argument,       NULL, 'P' },
#ifdef IP_HDRINCL
  { "sender-address",                   required_argument,       NULL, 'q' },
#endif
  { "out-iface-idx",                    required_argument,       NULL, 'Q' },
  { "local-traffic-direction",          no_argument,             NULL, 'r' },
  { "scan-cycle",                       required_argument,       NULL, 's' },
  { "sample-rate",                      required_argument,       NULL, 'S' },
  { "lifetime-timeout",                 required_argument,       NULL, 't' },
  { "flow-templ",                       required_argument,       NULL, 'T' },
  { "in-iface-idx",                     required_argument,       NULL, 'u' },
  { "flow-templ-id",                    required_argument,       NULL, 'U' },
  { "hash-size",                        required_argument,       NULL, 'w' },
  { "no-ipv6",                          no_argument,             NULL, 'W' },
  { "version",                          no_argument,             NULL, 'v' },
  { "flow-version",                     required_argument,       NULL, 'V' },
  { "min-flow-size",                    required_argument,       NULL, 'z' },

#ifdef HAVE_PF_RING
  { "cluster-id",                       required_argument,       NULL, 'Z' },
#endif

#ifdef HAVE_MYSQL
  { "mysql",                            required_argument,       NULL, '0' /* ignored */},
  { "mysql-skip-db-creation",           no_argument,             NULL, '0' /* ignored */},
#endif
  { "if-networks",                      required_argument,       NULL, '1' },
  { "count",                            required_argument,       NULL, '2' },
  { "collector-port",                   required_argument,       NULL, '3' },
#ifdef linux
  { "cpu-affinity",                     required_argument,       NULL, '4' },
#endif
  { "tunnel",                           no_argument,             NULL, '5' },
  /* Handled by the plugin */
  { "no-promisc",                       no_argument,             NULL, '6' },
  { "smart-udp-frags",                  no_argument,             NULL, '7' },
  { "ipsec-auth-data-len",              required_argument,       NULL, '8' },
  { "dump-stats",                       required_argument,       NULL, '9' },
  { "black-list",                       required_argument,       NULL, '!' },
  { "vlanid-as-iface-idx",              required_argument,       NULL, '@' },
  { "discard-unknown-flows",            required_argument,       NULL, '&' },
  { "pcap-file-list",                   required_argument,       NULL, '$' },
  { "csv-separator",                    required_argument,       NULL, '^' },
  { "city-list",                        required_argument,       NULL, ',' },
#ifdef HAVE_FASTBIT
  { "fastbit",                          required_argument,       NULL, '[' },
  { "fastbit-rotation",                 required_argument,       NULL, ']' },
  { "fastbit-template",                 required_argument,       NULL, '(' },
#ifndef WIN32
  { "fastbit-index",                    required_argument,       NULL, ')' },
#endif
  { "fastbit-exec",                     required_argument,       NULL, '#' },
#endif

  { "dont-drop-privileges",             no_argument,             NULL, '\\' },
  { "bi-directional",                   no_argument,             NULL, '{' },
  { "account-l2",                       no_argument,             NULL, '}' },
  { "dump-metadata",                    required_argument,       NULL, '=' },
  { "event-log",                        required_argument,       NULL, '+' },

  { "unprivileged-user",                required_argument,       NULL, 244 },
  { "disable-cache",                    no_argument,             NULL, 245 },
  { "fake-capture",                     no_argument,             NULL, 246 },
  { "quick-mode",                       no_argument,             NULL, 247 },
  { "performance",                      no_argument,             NULL, 248 },

  /*
    Options for plugins. These options are not handled by the main
    program but it's important to have them defined here otherwise we
    get a warning from the probe
  */
  { "log",                              required_argument, NULL, 249 /* dummy */ },
  { "nflite",                           required_argument, NULL, 250 /* dummy */ },
  { "dont-hash-cookies",                no_argument,       NULL, 251 /* dummy */ },
  { "dont-nest-dump-dirs",              no_argument,       NULL, 251 /* dummy */ },
  { "max-http-log-lines",               required_argument, NULL, 252 /* dummy */ },
  { "http-dump-dir",                    required_argument, NULL, 252 /* dummy */ },
  { "http-exec-cmd",                    required_argument, NULL, 252 /* dummy */ },
#ifdef HAVE_YASSL
  { "ssl-config-file",                  required_argument, NULL, 252 /* dummy */ },
#endif

  { "max-mysql-log-lines",              required_argument, NULL, 252 /* dummy */ },
  { "mysql-dump-dir",                   required_argument, NULL, 252 /* dummy */ },
  { "dns-exec-cmd",                     required_argument, NULL, 252 /* dummy */ },

  { "oracle-dump-dir",                  required_argument, NULL, 252 /* dummy */ },
  { "oracle-exec-cmd",                  required_argument, NULL, 252 /* dummy */ },
  { "max-oracle-log-lines",             required_argument, NULL, 252 /* dummy */ },

  { "dns-dump-dir",                     required_argument, NULL, 252 /* dummy */ },
  { "ntop-ng",                          required_argument, NULL, 252 /* dummy */ },
  { "bgp-port",                         required_argument, NULL, 252 /* dummy */ },
  { "nprobe-version",                   required_argument, NULL, 252 /* dummy */ },
  { "gtp-dump-dir",                     required_argument, NULL, 252 /* dummy */ },
  { "gtp-exec-cmd",                     required_argument, NULL, 252 /* dummy */ },
  { "radius-dump-dir",                  required_argument, NULL, 252 /* dummy */ },
  { "radius-exec-cmd",                  required_argument, NULL, 252 /* dummy */ },
  { "l7-rules",                         required_argument, NULL, 252 /* dummy */ },

  { "debug",                            no_argument,       NULL, 254 },

  /* End of probe options */
  { NULL,                               no_argument,       NULL,  0 }
};

/* ****************************************************** */

void printPcapStats(pcap_t *pcapPtr) {
  struct pcap_stat pcapStat;

  if(readOnlyGlobals.fakePacketCapture) return;

  if(pcap_stats(pcapPtr, &pcapStat) >= 0) {
    u_long rcvd_diff, drop_diff;
    char msg[256];

    /* Some pcap implementations resetthe stats at each call */
    if(pcapStat.ps_recv >= readWriteGlobals->last_ps_recv) {
      rcvd_diff = pcapStat.ps_recv - readWriteGlobals->last_ps_recv;
      drop_diff = pcapStat.ps_drop - readWriteGlobals->last_ps_drop;
    } else {
      rcvd_diff = pcapStat.ps_recv, drop_diff = pcapStat.ps_drop;
    }

    /* traceEvent(TRACE_ERROR, "[%u][%u]\n", pcapStat.ps_recv, pcapStat.ps_drop); */

    snprintf(msg, sizeof(msg), "Packet stats (pcap): "
	     "%u/%u pkts rcvd/dropped [%.1f%%] [Last %lu/%lu pkts rcvd/dropped]",
	     pcapStat.ps_recv - pcapStat.ps_drop, pcapStat.ps_drop,
	     (pcapStat.ps_recv > 0) ? ((float)(pcapStat.ps_drop*100)/(float)pcapStat.ps_recv) : 0,
	     rcvd_diff, drop_diff);

    traceEvent(TRACE_INFO, "%s", msg);
    if(drop_diff > 0) dumpLogEvent(packet_drop, severity_warning, msg);

    readWriteGlobals->last_ps_recv = pcapStat.ps_recv, readWriteGlobals->last_ps_drop = pcapStat.ps_drop;
  } else {
#ifdef DEBUG
    traceEvent(TRACE_WARNING, "Unable to read pcap statistics: %s",
	       pcap_geterr(pcapPtr));
#endif
  }
}

/* ****************************************************** */

static void printCaptureStats(void) {
#ifdef HAVE_PF_RING
  printPfRingStats();
#else
  if(readOnlyGlobals.pcapPtr != NULL)
    printPcapStats(readOnlyGlobals.pcapPtr);
#endif
}

/* ****************************************************** */

#ifndef WIN32

void reloadCLI(int signo) {
  traceEvent(TRACE_NORMAL, "Received signal %d: reloading CLI options", signo);

  parseOptions(argc_, argv_, 1);
}

/* ****************************************************** */

void cleanup(int signo) {
  static u_char statsPrinted = 0;

  if(!readOnlyGlobals.nprobe_up) exit(0);

  if(!statsPrinted) {
    statsPrinted = 1;
    printCaptureStats();
  }

  shutdown_nprobe();
  /* exit(0); */
}
#endif

/* ****************************************************** */

#ifndef WIN32
void brokenPipe(int signo) {
#ifdef DEBUG
  traceEvent(TRACE_WARNING, "Broken pipe (socket %d closed) ?\n", currSock);
#endif
  signal(SIGPIPE, brokenPipe);
}
#endif

/* ****************************************************** */

inline void  updateThreadPacketStats(u_short pktLen, u_short thread_id) {
  readWriteGlobals->accumulateStats[thread_id].pkts++, readWriteGlobals->accumulateStats[thread_id].bytes += pktLen;
  readWriteGlobals->currentPkts[thread_id]++, readWriteGlobals->currentBytes[thread_id] += pktLen;
}

/* ****************************************************** */

static void deepPacketDecode(u_short thread_id,
			     int packet_if_idx /* -1 = unknown */,
			     struct pcap_pkthdr *h, const u_char *p,
			     u_int8_t sampledPacket,
			     u_short numPkts, int input_index, int output_index,
			     u_int32_t flow_sender_ip,
			     u_int32_t packet_hash) {
  struct eth_header *ehdr = NULL;
  u_int caplen = h->caplen, length = h->len, offset = 0;
  u_short eth_type, off = 0;
  u_int8_t tcpFlags = 0, proto = 0;
  u_int8_t icmp_type = 0, icmp_code = 0;
  u_int32_t tunnel_id = 0;
  u_int32_t  tcpSeqNum = 0, tcpAckNum = 0;
  struct ip *ip = NULL;
  struct ip6_hdr *ipv6 = NULL;
  struct ip6_ext *ipv6ext = NULL;
  struct tcphdr *tp;
  struct udphdr *up;
  struct icmp_hdr *icmpPkt;
  u_int16_t payload_shift = 0;
  int originalPayloadLen = 0, payloadLen = 0; /* Do not set it to unsigned */
  IpAddress src, dst;
  IpAddress untunneled_src, untunneled_dst;
  u_int16_t untunneled_sport = 0, untunneled_dport = 0, ip_offset = 0, gtp_offset = 0;
  u_int8_t untunneled_proto = 0;
  u_short numFragments = 0;
  u_int ehshift = 0;
  ticks when;

#ifdef DEBUG
  traceEvent(TRACE_INFO, ".");
#endif

  /*
    FIX: when packet_hash > 0 check if the packet isn't fragmented
  */
  if(unlikely(readOnlyGlobals.tracePerformance)) when = getticks();

  updateThreadPacketStats(h->len, thread_id);

  if(unlikely(readWriteGlobals->stopPacketCapture)) return;

  if(unlikely(readOnlyGlobals.initialSniffTime.tv_sec == 0)) {
    /* Set it with the first incoming packet */
    memcpy(&readOnlyGlobals.initialSniffTime, &h->ts, sizeof(struct timeval));
  }

  readWriteGlobals->now = h->ts.tv_sec;

  if(likely(caplen >= sizeof(struct eth_header))) {
    u_int plen, hlen = 0, ip_len = 0;
    u_short sport, dport, numMplsLabels = 0, tcp_len;
    u_char mplsLabels[MAX_NUM_MPLS_LABELS][MPLS_LABEL_LEN];
    u_int32_t null_type;
    struct ppp_header ppphdr;

    // traceEvent(TRACE_INFO, "Datalink: %d", datalink);

    switch(readOnlyGlobals.datalink) {
    case DLT_ANY: /* Linux 'any' device */
      eth_type = DLT_ANY;
      break;
    case DLT_RAW: /* Raw packet data */
      if(((p[0] & 0xF0) >> 4) == 4)
	eth_type = ETHERTYPE_IP;
      else
	eth_type = ETHERTYPE_IPV6;
      ehshift = 0;
      break;
    case DLT_NULL: /* loopaback interface */
      ehshift = 4;
      memcpy(&null_type, p, sizeof(u_int32_t));
      //null_type = ntohl(null_type);
      /* All this crap is due to the old little/big endian story... */
      /* FIX !!!! */
      switch(null_type) {
      case BSD_AF_INET:
	eth_type = ETHERTYPE_IP;
	break;
      case BSD_AF_INET6_BSD:
      case BSD_AF_INET6_FREEBSD:
      case BSD_AF_INET6_DARWIN:
	eth_type = ETHERTYPE_IPV6;
	break;
      default:
	return; /* Any other non IP protocol */
      }
      break;
    case DLT_PPP:
      memcpy(&ppphdr, p, sizeof(struct ppp_header));
      if(ntohs(ppphdr.proto) == 0x0021 /* IP */)
	eth_type = ETHERTYPE_IP, ehshift = sizeof(struct ppp_header);
      else
	return;
      break;
    default:
      ehshift = sizeof(struct eth_header);
      ehdr = (struct eth_header*)p;
      eth_type = ntohs(ehdr->ether_type);
      break;
    }

    switch(eth_type) {
    case ETHERTYPE_IP:
    case ETHERTYPE_IPV6:
    case ETHERTYPE_VLAN:
    case ETHERTYPE_MPLS:
    case ETHERTYPE_PPPoE:
    case DLT_NULL:
    case DLT_ANY:
    case 16385 /* MacOSX loopback */:
    case 16390 /* MacOSX loopback */:
      {
	u_short vlanId = 0, outerVlanId = 0;
	u_int estimatedLen = 0;

	if(likely((eth_type == ETHERTYPE_IP) || (eth_type == ETHERTYPE_IPV6))) {
	  if((ehshift == 0) && (readOnlyGlobals.datalink != DLT_RAW)) /* still not set (used to handle the DLT_NULL case) */
	    ehshift = sizeof(struct eth_header);
	} else if(eth_type == ETHERTYPE_MPLS) {
	  char bos; /* bottom_of_stack */

	  memset(mplsLabels, 0, sizeof(mplsLabels));
	  bos = 0;
	  while(bos == 0) {
	    memcpy(&mplsLabels[numMplsLabels], p+ehshift, MPLS_LABEL_LEN);

	    bos = (mplsLabels[numMplsLabels][2] & 0x1), ehshift += 4, numMplsLabels++;
	    if((ehshift > caplen) || (numMplsLabels >= MAX_NUM_MPLS_LABELS))
	      return; /* bad packet */
	  }
	  eth_type = ETHERTYPE_IP;
	} else if(eth_type == ETHERTYPE_PPPoE) {
	  eth_type = ETHERTYPE_IP, ehshift += 8;
	} else if(eth_type == ETHERTYPE_VLAN) {
	  Ether80211q *qType;

	  while(eth_type == ETHERTYPE_VLAN) {
	    qType = (Ether80211q*)(p+ehshift);
	    vlanId = ntohs(qType->vlanId) & 0xFFF;
	    if(outerVlanId == 0) outerVlanId = vlanId;
	    eth_type = ntohs(qType->protoType);
	    ehshift += sizeof(Ether80211q);
	    /* printf("VlanId: %d\n", vlanId); <<<== NOT USED YET */
	  }

	  if(eth_type == 0x0800) {
	    /* Sanity check */

	    if(p[ehshift] == 0x60)
	      eth_type = ETHERTYPE_IPV6;
	  }
	} else if(eth_type == DLT_ANY) {
	  ehshift += sizeof(AnyHeader);
	  eth_type = ntohs(((AnyHeader*)p)->protoType);
	} else
	  ehshift += NULL_HDRLEN;

      parse_ip:
	ip_offset= ehshift;
	if(likely(eth_type == ETHERTYPE_IP)) {
	  u_short ip_ip_len;

	  ip = (struct ip*)(p+ehshift);
	  if(ip->ip_v != 4) return; /* IP v4 only */

	  /* blacklist check */
	  if(unlikely(readOnlyGlobals.numBlacklistNetworks > 0)) {
	    if(isBlacklistedAddress(&ip->ip_src)
	       || isBlacklistedAddress(&ip->ip_dst))
	      return;
	  }

	  ip_ip_len = htons(ip->ip_len);

	  ip_len = ((u_short)ip->ip_hl * 4);
	  estimatedLen = ehshift + ip_ip_len;
	  hlen = ip_len;
	  payloadLen = htons(ip->ip_len)-ip_len;

	  if(readOnlyGlobals.roundPacketLenWithIPHeaderLen)
	    length = estimatedLen;

	  src.ipVersion = 4, dst.ipVersion = 4;
	  if(unlikely(readOnlyGlobals.ignoreIP
		      || (readOnlyGlobals.setAllNonLocalHostsToZero
			  && (readOnlyGlobals.numLocalNetworks > 0)
			  && (!isLocalAddress(&ip->ip_src)))))
	    src.ipType.ipv4 = 0; /* 0.0.0.0 */
	  else
	    src.ipType.ipv4 = ntohl(ip->ip_src.s_addr);

	  if(unlikely(readOnlyGlobals.ignoreIP
		      || (readOnlyGlobals.setAllNonLocalHostsToZero
			  && (readOnlyGlobals.numLocalNetworks > 0)
			  && (!isLocalAddress(&ip->ip_dst)))))
	    dst.ipType.ipv4 = 0; /* 0.0.0.0 */
	  else
	    dst.ipType.ipv4 = ntohl(ip->ip_dst.s_addr);

	  proto = ip->ip_p;
	  off = ntohs(ip->ip_off) & 0x3fff;
	  numFragments = off ? 1 : 0;
	} else if(eth_type == ETHERTYPE_IPV6) {
	  u_short ipv6_ip_len;

	  if(unlikely(readOnlyGlobals.disableIPv6)) return;

	  ipv6 = (struct ip6_hdr*)(p+ehshift);
	  if(((ipv6->ip6_vfc >> 4) & 0x0f) != 6) return; /* IP v6 only */

	  ipv6_ip_len = htons(ipv6->ip6_plen);
	  estimatedLen = sizeof(struct ip6_hdr)+ehshift+ipv6_ip_len;

	  if(readOnlyGlobals.roundPacketLenWithIPHeaderLen)
	    length = estimatedLen;

	  hlen = sizeof(struct ip6_hdr);
	  src.ipVersion = 6, dst.ipVersion = 6;
	  payloadLen = ipv6_ip_len - hlen;

	  /* FIX: blacklist check for IPv6 */

	  /* FIX: isLocalAddress doesn't work with IPv6 */
	  if(unlikely(readOnlyGlobals.ignoreIP))
	    memset(&src.ipType.ipv6, 0, sizeof(struct in6_addr));
	  else
	    memcpy(&src.ipType.ipv6, &ipv6->ip6_src, sizeof(struct in6_addr));

	  if(unlikely(readOnlyGlobals.ignoreIP))
	    memset(&dst.ipType.ipv6, 0, sizeof(struct in6_addr));
	  else
	    memcpy(&dst.ipType.ipv6, &ipv6->ip6_dst, sizeof(struct in6_addr));

	  proto = ipv6->ip6_nxt; /* next header (protocol) */

	  if(proto == 0) {
	    /* IPv6 hop-by-hop option */

	    ipv6ext = (struct ip6_ext*)(p+ehshift+40);
	    hlen += (ipv6ext->ip6e_len+1)*8;
	    proto = ipv6ext->ip6e_nxt;
	  }
	} else
	  return; /* Anything else that's not IPv4/v6 */

	originalPayloadLen = payloadLen;
	plen = length-ehshift;
	if(caplen > estimatedLen) caplen = estimatedLen;
	payloadLen -= (estimatedLen-caplen);

	sport = dport = 0; /* default */
	offset = ehshift+hlen;

	if(unlikely(readOnlyGlobals.tunnel_mode)) {
	  switch(proto) {
	  case IPPROTO_ESP:
	    /* http://www.unixwiz.net/techtips/iguide-ipsec.html */
	    if(payloadLen > readOnlyGlobals.ipsec_auth_data_len) {
	      proto = p[offset+payloadLen-readOnlyGlobals.ipsec_auth_data_len-1];
	      offset += 8;
	    }
	    break;

	  case IPPROTO_GRE:
	    {
	      struct gre_header *gre;

	      gre = (struct gre_header*)&p[offset];
	      gre->flags_and_version = ntohs(gre->flags_and_version);
	      gre->proto = ntohs(gre->proto);

	      offset += sizeof(struct gre_header);
	      if(gre->flags_and_version & GRE_HEADER_CHECKSUM) offset += 4;
	      if(gre->flags_and_version & GRE_HEADER_ROUTING)  offset += 4;
	      if(gre->flags_and_version & GRE_HEADER_KEY)      offset += 4;
	      if(gre->flags_and_version & GRE_HEADER_SEQ_NUM)  offset += 4;

	      eth_type = gre->proto;

	      if(eth_type == 0x8881 /* CDMA2000 */) {
		offset++; /* PPP in HDLC-Like Framing */
		memcpy(&ppphdr, &p[offset], sizeof(struct ppp_header));
		if(ntohs(ppphdr.proto) == 0x0021 /* IP */)
		  eth_type = ETHERTYPE_IP;

		ehshift = sizeof(struct ppp_header)+offset;
	      } else
		ehshift = offset;

	      memcpy(&untunneled_src, &src, sizeof(IpAddress)), memcpy(&untunneled_dst, &dst, sizeof(IpAddress));
	      untunneled_proto = proto, untunneled_sport = sport, untunneled_dport = dport;
	      goto parse_ip;
	      break;
	    }
	  }
	}

	switch(proto) {
	case IPPROTO_TCP:
	  if(unlikely(plen < (hlen+sizeof(struct tcphdr)))) return; /* packet too short */
	  tp = (struct tcphdr*)(p+offset);
	  if(likely(!readOnlyGlobals.ignorePorts)) sport = ntohs(tp->th_sport), dport = ntohs(tp->th_dport);
	  tcpFlags = tp->th_flags, tcpSeqNum = ntohl(tp->th_seq), tcpAckNum = ntohl(tp->th_ack);
	  tcp_len = (tp->th_off * 4);
	  payloadLen -= tcp_len, originalPayloadLen -= tcp_len;
	  if(likely(payloadLen > 0))
	    payload_shift = offset+tcp_len;
	  else {
	    payloadLen    = 0;
	    payload_shift = 0;
	  }
	  break;

	case IPPROTO_UDP:
	  if(unlikely(plen < (hlen+sizeof(struct udphdr)))) return; /* packet too short */
	  up = (struct udphdr*)(p+offset);
	  if(likely(!readOnlyGlobals.ignorePorts)) sport = ntohs(up->uh_sport), dport = ntohs(up->uh_dport);
	  originalPayloadLen = payloadLen = ntohs(up->uh_ulen)-sizeof(struct udphdr);
	  if(likely(payloadLen > 0))
	    payload_shift = offset+sizeof(struct udphdr);
	  else {
	    payloadLen    = 0;
	    payload_shift = 0;
	  }

	  if(unlikely((readOnlyGlobals.tunnel_mode || readOnlyGlobals.enableGtpPlugin)
		      && (payloadLen > sizeof(struct gtp_header)))) {
	    if((sport == GTP_DATA_PORT) || (dport == GTP_DATA_PORT)) {
	      struct gtp_header *gtp = (struct gtp_header*)&p[payload_shift];
	      u_int gtp_header_len = 8 /* min size of struct gtp_header */;

	      if(((gtp->flags & 0x30) == 0x30) /* GTPv1 */
		 && (ntohs(gtp->total_length) >= (payloadLen-gtp_header_len))) {
		/* gtp_offset = payload_shift; */ /* <<= do not set it for user-data */
		tunnel_id = ntohl(gtp->tunnel_id);

		/* Now compute gtp_header_len precisely */
		if(gtp->flags & 0x04) gtp_header_len += 1; /* next_ext_header is present */
		if(gtp->flags & 0x02) gtp_header_len += 4; /* sequence_number is present (it also includes next_ext_header and pdu_number) */
		if(gtp->flags & 0x01) gtp_header_len += 1; /* pdu_number is present */

		payload_shift += gtp_header_len;
		ehshift = payload_shift;

		if(p[payload_shift] == 0x60)
		  eth_type = ETHERTYPE_IPV6;
		else
		  eth_type = ETHERTYPE_IP;

		memcpy(&untunneled_src, &src, sizeof(IpAddress)), memcpy(&untunneled_dst, &dst, sizeof(IpAddress));
		untunneled_proto = proto, untunneled_sport = sport, untunneled_dport = dport;
		goto parse_ip;
	      }
	    } else if((sport == GTP_CONTROL_PORT) || (dport == GTP_CONTROL_PORT)) {
	      struct gtp_header *gtp = (struct gtp_header*)&p[payload_shift];
	      u_int gtp_header_len = 8 /* min size of struct gtp_header */;

	      if(((gtp->flags & 0x30) == 0x30) /* GTPv1 */
		 && (ntohs(gtp->total_length) >= (payloadLen-gtp_header_len))) {
		gtp_offset = payload_shift;
	      }
	    }
	  }

	  if(unlikely(readOnlyGlobals.enable_debug)) {
	    if((payloadLen > 0)
	       && (numFragments == 0) && (off == 0) /* Do not process fragmented packets */
	       && ((dport == 2055)
		   || (dport == 2057)
		   || (dport == 9999)
		   || (dport == 3000)
		   || (dport == 6000)
		   || (dport == 9996)
		   || (dport == 15003)
		   )) {
	      /* traceEvent(TRACE_NORMAL, "Dissecting flow packets (%d bytes)", payloadLen); */
#if 0
	      int begin = 70;

	      traceEvent(TRACE_NORMAL, "%02X %02X %02X %02X %02X %02X",
			 p[payload_shift+begin] & 0xFF, p[payload_shift+begin+1] & 0xFF,
			 p[payload_shift+begin+2] & 0xFF, p[payload_shift+begin+3] & 0xFF,
			 p[payload_shift+begin+4] & 0xFF, p[payload_shift+begin+5] & 0xFF);
#endif

	      dissectNetFlow(0, (char*)&p[payload_shift], payloadLen);
	      return;
	    }
	  }

	  break;
	case IPPROTO_ICMP:
	case IPPROTO_ICMPV6:
	  if(plen < (hlen+sizeof(struct icmp_hdr))) return; /* packet too short */
	  icmpPkt = (struct icmp_hdr*)(p+offset);
	  payloadLen = caplen - offset - sizeof(struct icmp_hdr);
	  //traceEvent(TRACE_ERROR, "[icmp_type=%d][icmp_code=%d]", icmpPkt->icmp_type, icmpPkt->icmp_code);
	  if(!(readOnlyGlobals.ignorePorts || readOnlyGlobals.ignorePorts)) {
	    if(readOnlyGlobals.usePortsForICMP)
	      sport = 0, dport = (icmpPkt->icmp_type * 256) + icmpPkt->icmp_code;
	  }

	  icmp_type = icmpPkt->icmp_type, icmp_code = icmpPkt->icmp_code;
	  if(payloadLen > 0) {
	    payload_shift = offset;

	    if(proto == IPPROTO_ICMP)
	      payload_shift += sizeof(struct icmp_hdr);
	    else
	      payload_shift += 64; /* ICMPv6 */
	  } else {
	    payloadLen    = 0;
	    payload_shift = 0;
	  }
	  break;

	default:
	  payloadLen = 0;
	}

	/* ************************************************ */

	/* Is this a fragment ?
	   NOTE: IPv6 doesn't have the concept of fragments
	*/
	if(unlikely(readOnlyGlobals.handleFragments && (numFragments > 0))) {
	  u_short fragmentOffset = (off & 0x1FFF)*8, fragmentId = ntohs(ip->ip_id);
	  u_short fragment_list_idx = (src.ipType.ipv4 + dst.ipType.ipv4) % NUM_FRAGMENT_LISTS;
	  IpV4Fragment *list, *prev = NULL;

	  if((readOnlyGlobals.smart_udp_frags_mode == 0) || (proto != IPPROTO_UDP)) {
	    pthread_rwlock_wrlock(&readWriteGlobals->fragmentMutex[fragment_list_idx]);
	    list = readWriteGlobals->fragmentsList[fragment_list_idx];

	    while(list != NULL) {
	      if((list->src == src.ipType.ipv4)
		 && (list->dst == dst.ipType.ipv4)
		 && (list->fragmentId == fragmentId))
		break;
	      else {
		if((h->ts.tv_sec-list->firstSeen) > 30 /* sec */) {
		  /* Purge expired fragment */
		  IpV4Fragment *next = list->next;

		  if(prev == NULL)
		    readWriteGlobals->fragmentsList[fragment_list_idx] = next;
		  else
		    prev->next = next;

		  free(list);
		  readWriteGlobals->fragmentListLen[fragment_list_idx]--;
		  list = next;
		} else {
		  prev = list;
		  list = list->next;
		}
	      }
	    }

	    if(list == NULL) {
	      /* Fragment not found */
	      IpV4Fragment *frag = (IpV4Fragment*)malloc(sizeof(IpV4Fragment));

	      /* We have enough memory */
	      if(frag != NULL) {
		memset(frag, 0, sizeof(IpV4Fragment));
		frag->next = readWriteGlobals->fragmentsList[fragment_list_idx];
		readWriteGlobals->fragmentsList[fragment_list_idx] = frag;
		frag->src = src.ipType.ipv4, frag->dst = dst.ipType.ipv4;
		frag->fragmentId = fragmentId;
		frag->firstSeen = h->ts.tv_sec;
		list = frag, prev = NULL;;
		readWriteGlobals->fragmentListLen[fragment_list_idx]++;
	      } else
		traceEvent(TRACE_ERROR, "Not enough memory?");
	    }

	    if(list != NULL) {
	      if(fragmentOffset == 0)
		list->sport = sport, list->dport = dport;

	      list->len += plen, list->numPkts++;

	      if(!(off & IP_MF)) {
		/* last fragment->we know the total data size */
		IpV4Fragment *next = list->next;
		sport = list->sport, dport = list->dport;
		plen = list->len, numPkts = list->numPkts;

		/* We can now free the fragment */
		if(prev == NULL)
		  readWriteGlobals->fragmentsList[fragment_list_idx] = next;
		else
		  prev->next = next;

		readWriteGlobals->fragmentListLen[fragment_list_idx]--;
		free(list);
		pthread_rwlock_unlock(&readWriteGlobals->fragmentMutex[fragment_list_idx]);
		numFragments = numPkts;
	      } else {
		pthread_rwlock_unlock(&readWriteGlobals->fragmentMutex[fragment_list_idx]);
		/* More fragments: we'll handle the packet later */
		return;
	      }
	    }
	  } else {
	    if(fragmentOffset > 0) {
	      /*
		Ignore fragments that do not have the initial
		fragmented packet info
	      */
	      return;
	    } else {
	      /*
		We use 2* because we want to be as precise as possible given
		that we have at least two fragments, we account twice the
		IP packt header
	      */
	      plen = ntohs(up->uh_ulen)+2*ip_len, numPkts = 2;
	    }
	  }
	}

	/* ************************************************ */

#ifdef DEBUG
	{
	  char buf[256], buf1[256];

	  printf("%2d) %s:%d -> %s:%d [len=%d][payloadLen=%d]\n",
		 ip->ip_p, _intoaV4(ip->ip_src.s_addr, buf, sizeof(buf)), sport,
		 _intoaV4(ip->ip_dst.s_addr, buf1, sizeof(buf1)), dport,
		 plen, payloadLen);
	}
#endif

	if(unlikely((src.ipVersion == 4) && (src.ipType.ipv4 == 0)
		    && (dst.ipType.ipv4 == 0) && (!(readOnlyGlobals.ignoreIP))))
	  return; /* Flow to skip */

	if(unlikely(readOnlyGlobals.tracePerformance)) {
	  ticks diff = getticks() - when;

	  pthread_rwlock_wrlock(&readOnlyGlobals.ticksLock);
	  readOnlyGlobals.decodeTicks += diff;
	  pthread_rwlock_unlock(&readOnlyGlobals.ticksLock);
	}

	if(unlikely(numPkts == 0)) {
	  traceEvent(TRACE_WARNING, "[%u] Internal error (zero packets)", pthread_self());
	} else {
	  u_int8_t ttl;

	  if(readOnlyGlobals.use_vlanId_as_ifId == outer_vlan)
	    vlanId = outerVlanId;

	  if(ip) ttl = ip->ip_ttl;
	  else if(ipv6) ttl = ipv6->ip6_ctlun.ip6_un1.ip6_un1_hlim;
	  else ttl = 0;

	  if(readOnlyGlobals.quick_mode) {
	    quickProcessFlowPacket(thread_id, packet_if_idx,
				   proto, ip_offset, numPkts, vlanId,
				   &src, sport, &dst, dport,
				   input_index, output_index,
				   (struct pcap_pkthdr*)h, (u_char*)p,
				   payload_shift, payloadLen,
				   originalPayloadLen,
				   packet_hash);	    
	  } else {
	    processFlowPacket(thread_id, packet_if_idx,
			      proto, numFragments, ip_offset, sampledPacket,
			      numPkts,
			      ip ? ip->ip_tos : 0, ttl,
			      vlanId, tunnel_id, gtp_offset,
			      ehdr, &src, sport, &dst, dport,
			      untunneled_proto,
			      &untunneled_src, untunneled_sport,
			      &untunneled_dst, untunneled_dport,
			      unlikely(readOnlyGlobals.accountL2Traffic) ? h->len : plen,
			      tcpFlags, tcpSeqNum, tcpAckNum,
			      icmp_type, icmp_code,
			      numMplsLabels, mplsLabels,
			      input_index, output_index,
			      (struct pcap_pkthdr*)h, (u_char*)p,
			      payload_shift, payloadLen,
			      originalPayloadLen, 0,
			      0, 0, 0, 0, flow_sender_ip,
			      packet_hash);
	  }
	}
      }
      break;

    default:
#ifdef DEBUG
      traceEvent(TRACE_WARNING, "Unknown ethernet type: 0x%X (%d)", eth_type, eth_type);
#endif
      readWriteGlobals->discardedPkts[thread_id]++;
      
#ifdef HAVE_PF_RING
      if(readOnlyGlobals.enableL7BridgePlugin && (packet_if_idx != -1))
	forwardPacket(packet_if_idx, (char*)p, h->caplen);
#endif

      break;
    }
  }
}

/* ****************************************************** */

inline u_int8_t isQueueFull(PacketQueue *queue) {
  if(queue->queueSlots[queue->insert_idx].packet_ready
     && (queue->insert_idx == queue->remove_idx))
    return(1);
  else
    return(0);
}

/* ****************************************************** */

void decodePacket(u_short thread_id,
		  int packet_if_idx /* -1 = unknown */,
		  struct pcap_pkthdr *h, const u_char *p,
		  u_int8_t sampledPacket,
		  u_short numPkts, int input_index, int output_index,
		  u_int32_t flow_sender_ip,
		  u_int32_t packet_hash) {

  /* Sanity check */
  if(unlikely((h->ts.tv_sec < 0) || (h->ts.tv_usec < 0))) {
    traceEvent(TRACE_WARNING, "Invalid timestamp: %lu.%lu", h->ts.tv_sec, h->ts.tv_usec);
    return; /* We ignore this packet */
  } else if(unlikely((h->caplen > h->len) || (h->len > 16384))) {
    traceEvent(TRACE_WARNING, "Invalid packet length: [len=%lu][caplen=%lu]", h->len, h->caplen);
    traceEvent(TRACE_WARNING, "Please disable LRO/GRO on your NIC (ethtool -k <NIC>)");
    return; /* We ignore this packet */
  }

  if(packet_hash) {
    if(readOnlyGlobals.numProcessThreads > 1) {
      u_int queue_id = packet_hash >> 2; /* shuffle data a bit */
      PacketQueue *queue = &readWriteGlobals->packetQueues[queue_id % readOnlyGlobals.numProcessThreads];
      QueuedPacket *slot;

      while(isQueueFull(queue)) {
	if(thread_id > 0) traceEvent(TRACE_WARNING, "Queue %d is full [packet_hash=%u]", thread_id, packet_hash);
	if(readWriteGlobals->shutdownInProgress)
	  return;
	else
	  usleep(1);
      }

      /* Copy the packet in queue */
      slot = &queue->queueSlots[queue->insert_idx];
      slot->packet_if_idx = packet_if_idx;
      memcpy(&slot->h, h, sizeof(struct pcap_pkthdr)), memcpy(slot->p, p, h->caplen);
      slot->packet_ready = 1, slot->packet_hash = packet_hash;
      queue->insert_idx = (queue->insert_idx + 1) % DEFAULT_QUEUE_CAPACITY, queue->num_insert++;
      signalCondvar(&queue->dequeue_condvar, 0);
      return;
    }
  }

  /* We need to use the default path */
  deepPacketDecode(thread_id, packet_if_idx,
		   h, p, sampledPacket,
		   numPkts, input_index, output_index,
		   flow_sender_ip, packet_hash);
}

/* ****************************************************** */

void dummyProcessPacket(u_char *_deviceId,
			const struct pcap_pkthdr *h,
			const u_char *p) {
  // traceEvent(TRACE_NORMAL, "Got %d bytes packet", h->len);
  decodePacket(0 /* fake threadId */,
	       -1 /* input interface id */,
	       (struct pcap_pkthdr*)h, p,
	       0 /* sampledPacket */, 1 /* numPkts */,
	       NO_INTERFACE_INDEX, NO_INTERFACE_INDEX,
	       0 /* flow_sender_ip */,
	       0 /* packet hash */);
}

/* ****************************************************** */

void freeHostHash(void) {
  if(readOnlyGlobals.enableHostStats) {
    traceEvent(TRACE_INFO, "MISSING implement freeHostHash()");
  }
}

/* ****************************************************** */

static void msecSleep(u_int msSleep) {
#ifndef WIN32
  struct timespec timeout;

  timeout.tv_sec = 0, timeout.tv_nsec = 1000000*msSleep;

  while((nanosleep(&timeout, &timeout) == -1) && (errno == EINTR))
    ; /* Do nothing */
#else
  waitForNextEvent(msSleep);
#endif
}

/* ****************************************************** */

/*
  From the tests carried on, the very best approach
  is to have a periodic thread that scans for expired
  flows.
*/
void* hashWalker(void* notused) {
  u_short sleep_time, msSleep = 100;
  long idx;

  /* Wait until all the data structures have been allocated */
  while(readWriteGlobals->theFlowHash[readOnlyGlobals.numProcessThreads-1] == NULL) ntop_sleep(1);

  /* Align to the scan cycle */
  sleep_time = readOnlyGlobals.scanCycle - (time(NULL) % readOnlyGlobals.scanCycle);

  if(readOnlyGlobals.traceMode == 2)
    traceEvent(TRACE_INFO, "Sleeping %d sec before walking hash for the first time", sleep_time);
  ntop_sleep(sleep_time);

  while((readWriteGlobals->shutdownInProgress == 0) && (readWriteGlobals->stopPacketCapture == 0)) {
    int i;

    printStats(0);

    for(i=0; i<readOnlyGlobals.numProcessThreads; i++) idleThreadTask(i);
    checkNetFlowExport(0);
    ntop_sleep(readOnlyGlobals.scanCycle);
  } /* while */

  traceEvent(TRACE_INFO, "Hash walker thread terminated");
  return(NULL);
}

/* ****************************************************** */

#if defined(HAVE_LICENSE) || defined(WIN32)
int isGoodLicense() {
  extern int verify_license(char * version, char *license_path, int kind, char *out_buf, int out_buf_len);
  char out_buf[512];

  if((verify_license(version, LICENSE_FILE_NAME, 2, out_buf, sizeof(out_buf)) != 0)
     && (verify_license(version, "/etc/"LICENSE_FILE_NAME, 2, out_buf, sizeof(out_buf)) != 0)) {
    return(0);
  } else
    return(1);
}
#endif

/* ****************************************************** */

void probeVersion(void) {
  char *sysId;

  printf("\nWelcome to nprobe v.%s (%s) for %s\n"
	 "%s\n"
	 "Copyright 2002-12 by Luca Deri <deri@ntop.org>\n",
	 version, nprobe_revision, osName,
#ifdef HAVE_PF_RING
	 "with native PF_RING acceleration.\n");
#else
  "");
#endif

#if defined(HAVE_LICENSE) || defined(WIN32)
  printf("\nSystemID: %s\n", (sysId = getSystemId()));
  free(sysId);

  if(!isGoodLicense()) {
    printf("WARNING: Invalid or missing nProbe license ("
#ifndef WIN32
	   "/etc/"
#endif
	   LICENSE_FILE_NAME")\n");
  } else {
    printf("Valid nProbe license found\n");
  }
#endif
}

/* ******************************************************** */

void usage(void) {
  char buf[16];
  char *prot_long_str[] = { IPOQUE_PROTOCOL_LONG_STRING };
  int i;

  initDefaults();
  probeVersion();

  printf("\nUsage:\n");

  printf("nprobe -n <host:port|none> [-i <interface|dump file>] [-t <lifetime timeout>]\n"
	 "              [-d <idle timeout>] [-l <queue timeout>] [-s <scan cycle>] [-N]\n"
	 "              [-p <aggregation>] [-f <filter>] [-a] [-b <level>]"
#ifndef WIN32
	 " [-G]"
	 " [-O <# threads>]"
#if defined(linux) || defined(__linux__)
	 " [-X]"
#endif
#endif
	 "\n              "
	 "[-P <path>] [-F <dump timeout>] [-D <format>] "
	 "\n              "
	 "[-u <in dev idx>] [-Q <out dev idx>]"
	 "\n              "
#ifndef WIN32
	 "[-I <probe name>] "
#endif
	 "[-v] [-w <hash size>] [-e <flow delay>] [-B <packet count>]\n"
	 "              [-z <min flow size>] [-M <max num flows>]"
	 "[-R <payload Len>]"
	 "\n              [-x <payload policy>] [-E <engine>] [-C <flow lock file>]"
	 "\n              [-m <min # flows>]"
#ifdef IP_HDRINCL
	 "[-q <host:port>]"
#endif
	 "\n              [-S <sample rate>] [-A <AS list>] [-g <PID file>]"
	 "\n              [-T <flow template>] [-U <flow template id>]"
	 "\n              [-o <v9 templ. export policy>] [-L <local nets>] [-c] [-r]"
	 "\n              [-1 <interface nets>] [-2 <number>] [-3 <port>] [-4] [-5 <port>] [-6]"
	 "\n              [-9 <path>] [--black-list <networks>] [--pcap-file-list <filename>]"
#ifdef HAVE_FASTBIT
	 "\n              [--fastbit <dir>] [--fastbit-rotation <mins>]"
	 "\n              [--fastbit-template <flow template>]"
#ifndef WIN32
	 " [--fastbit-index <flow template>]"
#endif
	 "\n              [--fastbit-exec <cmd>]"
#endif
#ifndef WIN32
	 " [--dont-drop-privileges]\n"
#endif
	 "\n\n"
	 );

  printf("[--collector|-n] <host:port|none>   | Address of the NetFlow collector(s).\n"
	 "                                    | Multiple collectors can be defined using\n"
         "                                    | multiple -n flags. In this case flows\n"
         "                                    | will be sent in round robin mode to\n"
         "                                    | all defined collectors if the -a flag\n"
	 "                                    | is used. Note that you can specify\n"
	 "                                    | both IPv4 and IPv6 addresses.\n"
         "                                    | If you specify none as value,\n"
         "                                    | no flow will be export; in this case\n"
	 "                                    | the -P parameter is mandatory.\n"
	 "                                    | Note that you can specify the protocol\n"
	 "                                    | used to send packets. Example:\n"
	 "                                    | udp://192.168.0.1:2055, tcp://10.1.2.3:2055\n");
#ifndef WIN32
  printf("[--interface|-i] <iface|pcap>       | Interface name from which packets are\n");
  printf("                                    | captured, or .pcap file (debug only)\n");
#else
  printf("[--interface|-i] <iface>            | Index or name of the interface from which\n");
  printf("                                    | packets are captured. Type -h to print\n");
  printf("                                    | all the know interfaces.\n");
#endif
  printf("[--lifetime-timeout|-t] <timeout>   | It specifies the maximum (seconds) flow\n"
	 "                                    | lifetime [default=%d]\n", readOnlyGlobals.lifetimeTimeout);
  printf("[--idle-timeout|-d] <timeout>       | It specifies the maximum (seconds) flow\n"
         "                                    | idle lifetime [default=%d]\n", readOnlyGlobals.idleTimeout);
  printf("[--queue-timeout|-l] <timeout>      | It specifies how long expired flows\n"
     	 "                                    | (queued before delivery) are emitted\n"
	 "                                    | [default=%d]\n", readOnlyGlobals.sendTimeout);
  printf("[--scan-cycle|-s <scan cycle>]      | It specifies how often (seconds) expired\n"
	 "                                    | flows are emitted [default=%d].\n"
	 "                                    | If -P is used, the scan cycle will be\n"
	 "                                    | set to the value of the -F parameter\n",
	 readOnlyGlobals.scanCycle);
  printf("[--aggregation|-p] <aggregation>    | It specifies the flow aggiregation level:\n"
	 "                                    | <VLAN Id>/<proto>/<IP>/<port>/<TOS>/<AS>\n"
	 "                                    | where each element can be set to 0=ignore\n"
	 "                                    | or 1=take care. Example \'-p 1/0/1/1/1/1\'\n"
	 "                                    | ignores the protocol, whereas\n"
	 "                                    | \'-p 0/0/1/0/0/0\' ignores everything\n"
	 "                                    | but the IP\n");
  printf("[--bpf-filter|-f] <BPF filter>      | BPF filter for captured packets\n"
	 "                                    | [default=no filter]\n");
  printf("[--all-collectors|-a]               | If several collectors are defined, this\n"
         "                                    | option gives the ability to send all\n"
         "                                    | collectors all the flows. If the flag is\n"
	 "                                    | omitted collectors are selected in\n"
	 "                                    | round robin.\n");
  printf("[--verbose|-b] <level>              | Verbose output:\n"
         "                                    | 0 - No verbose logging\n"
	 "                                    | 1 - Limited logging (traffic statistics)\n"
         "                                    | 2 - Full verbose logging\n");

#ifndef WIN32
  printf("[--daemon-mode|-G]                  | Start as daemon.\n");
#endif

  printf("[--num-threads|-O] <# threads>      | Number of packet fetcher threads\n"
	 "                                    | [default=%u]. Use 1 unless you know\n"
	 "                                    | what you're doing.\n",
	 readOnlyGlobals.numProcessThreads);
  printf("[--dump-path|-P] <path>             | Directory where dump files will\n"
	 "                                    | be stored.\n");
  printf("[--dump-frequency|-F] <dump timeout>| Dump files dump frequencey (sec).\n"
         "                                    | Default: %d\n", readOnlyGlobals.file_dump_timeout);
  printf("[--dump-format|-D] <format>         | <format>: flows are saved as:\n"
	 "                                    | b       : raw/uncompressed flows\n"
	 "                                    | B       : raw core flow fields (flow size: %u bytes)\n"
	 "                                    | t       : text flows\n"
#ifdef HAVE_SQLITE
	 "                                    | d       : SQLite\n"
#endif
	 "                                    | Example: -D b. Note: this flag has no\n"
	 "                                    | effect without -P.\n", 
	 (unsigned int)sizeof(FlowHashBucketCoreFields));
  printf("[--in-iface-idx|-u] <in dev idx>    | Index of the input device used in the\n");
  printf("                                    | emitted flows (incoming traffic). The default\n"
	 "                                    | value is %d. Use -1 as value to dynamically\n"
	 "                                    | set to the last two bytes of\n"
	 "                                    | the MAC address of the flow sender.\n",
	 (int16_t)readOnlyGlobals.inputInterfaceIndex);
  printf("[--out-iface-idx|-Q] <out dev idx>  | Index of the output device used in the\n");
  printf("                                    | emitted flows (outgoing traffic). The default\n"
	 "                                    | value is %d. Use -1 as value to dynamically\n"
	 "                                    | set to the last two bytes of\n"
	 "                                    | the MAC address of the flow receiver.\n",
	 (int16_t)readOnlyGlobals.outputInterfaceIndex);
  printf("[--vlanid-as-iface-idx] <mode>      | Use vlanId (or 0 if the traffic isn't tagged)\n"
	 "                                    | as interface index. Mode specifies in case of\n"
	 "                                    | stacked VLANs which vlanId to choose. Possible\n"
	 "                                    | values are 'inner' or 'outer'. Note that this option\n"
	 "                                    | superseedes the --in/out-iface-idx options\n");
  printf("[--discard-unknown-flows] <mode>    | In case you enable L7 proto detection (e.g. add %%L7_PROTO\n"
	 "                                    | to the template) this options enables you not to export\n"
	 "                                    | flows for which nDPI has not been able to detect the proto.\n"
         "                                    | Mode values:\n"
         "                                    | 0 - Export both known and unknown flows (default)\n"
	 "                                    | 1 - Export only known flows (discard flows with unknown protos)\n"
         "                                    | 2 - Export only unknown flows (discard flows with known protos)\n");
  printf("[--nprobe-version|-v]               | Prints the program version.\n");
  printf("[--flow-lock|-C] <flow lock>        | If the flow lock file is present no flows\n"
	 "                                    | are emitted. This facility is useful to\n"
	 "                                    | implement high availability by means of\n"
	 "                                    | a daemon that can create a lock file\n"
	 "                                    | when this instance is in standby.\n");
  printf("[--help|-h]                         | Prints this help.\n");
  printf("[--debug]                           | Enable debugging (development only).\n");
  printf("[--quick-mode]                      | Micro-nprobe.\n");
  printf("[--fake-capture]                    | Fake packet capture (development only).\n");
  printf("[--performance]                     | Enable performance tracing (development only).\n");

#ifndef WIN32
  printf("[--syslog|-I] <probe name>          | Log to syslog as <probe name>\n"
	 "                                    | [default=stdout]\n");
#endif
  printf("[--hash-size|-w] <hash size>        | Flows hash size [default=%d]\n",
	 readOnlyGlobals.flowHashSize);
  printf("[--no-ipv6|-W]                      | IPv6 packets/traffic will not be accounted.\n");
  printf("[--flow-delay|-e] <flow delay>      | Delay (in ms) between two flow\n"
	 "                                    | exports [default=%d]\n",
	 readOnlyGlobals.flowExportDelay);
  printf("[--count-delay|-B] <packet count>   | Send this many packets before\n"
	 "                                    | the -e delay [default=%d]\n",
	 readOnlyGlobals.packetFlowGroup);

  if(readOnlyGlobals.minFlowSize == 0)
    strcpy(buf, "unlimited");
  else
    sprintf(buf, "%u", readOnlyGlobals.minFlowSize);

  printf("[--min-flow-size|-z] <min flow size>| Minimum TCP flow size (in bytes).\n"
	 "                                    | If a TCP flow is shorter than the\n"
	 "                                    | specified size the flow is not\n"
	 "                                    | emitted [default=%s]\n", buf);
#ifdef HAVE_PF_RING
  printf("[--cluster-id] <cluster id>         | Specify the PF_RING clusterId on which\n"
	 "                                    | incoming packets will be bound.\n");
#endif

  printf("[--max-num-flows|-M] <max num flows>| Limit the number of active flows. This is\n"
         "                                    | useful if you want to limit the memory\n"
	 "                                    | or CPU allocated to nProbe in case of non\n"
	 "                                    | well-behaved applications such as\n"
	 "                                    | worms or DoS. [default=%u]\n",
	 readOnlyGlobals.maxNumActiveFlows);
  printf("[--netflow-engine|-E] <engine>      | Specify the engine type and id.\n"
	 "                                    | The format is engineType:engineId.\n"
	 "                                    | [default=%d:%d] where engineId is a\n"
	 "                                    | random number.\n",
	 readOnlyGlobals.engineType, readOnlyGlobals.engineId);
  printf("[--min-num-flows|-m] <min # flows>  | Minimum number of flows per packet\n"
	 "                                    | unless an expired flow is queued\n"
	 "                                    | for too long (see -l) [default=%d\n"
	 "                                    | for v5, dynamic for v9]\n",
	 V5FLOWS_PER_PAK);
  printf("[--sender-address|-q] <host:port>   | Specifies the address:port of the flow\n"
	 "                                    | sender. This optionis useful for hosts\n"
	 "                                    | with multiple interfaces or if flows\n"
	 "                                    | must be emitted from a static port\n");
  printf("[--sample-rate|-S] <pkt rate>:<flow rate>\n"
	 "                                    | Packet capture sampling rate and flow\n"
	 "                                    | sampling rate. If <pkt rate> starts with '@'\n"
	 "                                    | it means that nprobe will report the specified\n"
	 "                                    | sampling rate but will not sample itself\n"
	 "                                    | as incoming packets are already sampled\n"
	 "                                    | on the specified capture device at the\n"
	 "                                    | specified rate. Default: 1:1 [no sampling]\n");
  printf("[--as-list|-A] <AS list>            | GeoIP file containing the list of known ASs.\n"
	 "                                    | Example: GeoIPASNum.dat\n");
  printf("[--city-list] <City list>           | GeoIP file containing the city/IP mapping.\n"
	 "                                    | Example: GeoLiteCity.dat\n");
  printf("[--pid-file|-g] <PID file>          | Put the PID in the specified file\n");
  printf("[--flow-templ|-T] <flow template>   | Specify the NFv9 template (see below).\n");
  printf("[--flow-templ-id|-U] <templ. id>    | Specify the NFv9 template identifier\n"
	 "                                    | [default: %d]\n", readOnlyGlobals.idTemplate);
  printf("[--flow-version|-V] <version>       | NetFlow Version: 5=v5, 9=v9, 10=IPFIX\n");
  printf("[--flows-intra-templ|-o] <num>      | Specify how many flow pkts are exported\n"
	 "                                    | between template exports [default: %d]\n",
	 readOnlyGlobals.templatePacketsDelta);
  printf("[--local-networks|-L] <nets>        | Specify the list of local networks whose\n"
	 "                                    | format is <net>/<mask> (if multiple use comma).\n");
  printf("[--local-hosts-only|-c]             | All the IPv4 hosts outside the local\n"
	 "                                    | network lists will be set to 0.0.0.0\n"
	 "                                    | (-L must be specified before -c).\n"
	 "                                    | This reduces the load on the probe\n"
	 "                                    | instead of discarding flows on the\n"
	 "                                    | collector side.\n");
  printf("[--local-traffic-direction|-r]      | All the traffic going towards\n"
	 "                                    | the local networks (-L must also be\n"
	 "                                    | specified before -r) is assumed incoming\n"
	 "                                    | traffic all the rest is assumed outgoing\n"
	 "                                    | (see also -u and -Q).\n");
  printf("[--if-networks|-1] <nets>           | Specify the binding between interfaceId\n"
	 "                                    | and a network (see below).\n");
  printf("[--count|-2] <number>               | Capture a specified number of packets\n"
	 "                                    | and quit (debug only)\n");
  printf("[--collector-port|-3] <port>        | NetFlow/sFlow collector port for incoming flows\n");
#ifdef linux
  printf("[--cpu-affinity|-4] <CPU/Core Id>   | Binds this process to the specified CPU/Core\n"
	 "                                    | Note: the first available CPU corresponds to 0.\n");
#endif
  printf("[--tunnel|-5]                       | Compute flows on tunneled traffic rather than\n"
	 "                                    | on the external envelope\n");
  printf("[--no-promisc|-6]                   | Capture packets in non-promiscuous mode\n");
  printf("[--smart-udp-frags|-7]              | Ignore UDP fragmented packets with fragment offset\n"
	 "                                    | greater than zero, and compute the fragmented\n"
	 "                                    | packet length on the initial fragment header.\n");
  printf("[--ipsec-auth-data-len|-8] <len>    | Length of the authentication data of IPSec\n"
	 "                                    | in tunnel mode. If not set, IPSec will not be decoded\n");
  printf("[--dump-stats|-9] <path>            | Periodically dump traffic stats into the\n"
	 "                                    | specified file\n");
  printf("--black-list <networks>             | All the IPv4 hosts inside the networks\n"
         "                                    | black-list will be discarded.\n"
         "                                    | This reduces the load on the probe\n"
         "                                    | instead of discarding flows on the\n"
         "                                    | collector side.\n");
  printf("--pcap-file-list <filename>         | Specify a filename containing a list\n"
	 "                                    | of pcap files.\n"
	 "                                    | If you use this flag the -i option will be\n"
	 "                                    | ignored.\n");
  printf("--csv-separator <separator>         | Specify the separator for text files (see -P)\n"
	 "                                    | Default is '|' (pipe)\n");
#ifdef HAVE_FASTBIT
  printf("--fastbit <dir>                     | Base directory where FastBit files will be created.\n");
  printf("--fastbit-rotation <mins>           | Every <mins> minutes a new FastBit sub-directory is created\n"
	 "                                    | so that each directory contains at most <mins> minutes.\n"
	 "                                    | Default %d min(s).\n",
	 readOnlyGlobals.fastbit_mins_rotation);
  printf("--fastbit-template <flow template>  | Fields that will be dumped on FastBit partition. Its syntax\n"
	 "                                    | is the same as the -T flag. If this flag is not specified,\n"
	 "                                    | all the specified flow elements (-T) will be dumped.\n");
  printf("--fastbit-index <flow template>     | Index each directory containing FastBit files as soon as\n"
	 "                                    | the directory has been dumped. The flow template specifies\n"
	 "                                    | which columns will be indexed. Its syntax is the same as\n"
	 "                                    | the -T flag. This option requires that fbindex application\n"
	 "                                    | is installed or built. If this flag is not specified, all\n"
	 "                                    | columns will be indexed.\n");
  printf("--fastbit-exec <cmd>                | Execute the specified command after a directory has been\n"
         "                                    | dumped (and optionally indexed). The command must take an\n"
	 "                                    | argument that is the path to the directory just dumped.\n");
#endif
#ifndef WIN32
  printf("--dont-drop-privileges              | Do not drop privileges changing to user nobody\n");
#endif
  printf("--bi-directional                    | Force flows to be bi-directional. This option\n"
	 "                                    | is not supported by NetFlow V5 that by nature\n"
	 "                                    | supports only mono-directional flows\n");
  printf("--account-l2                        | NetFlow accounts IP traffic only, not counting\n"
	 "                                    | L2 headers. Using this option the L2 headers\n"
	 "                                    | are also accounted\n");
  printf("--dump-metadata <file>              | Dump flow metadata into the specified file\n"
	 "                                    | and quit\n");
  printf("--event-log <file>                  | Dump relevant activities into the specified log file\n");
  printf("--unprivileged-user <name>          | Use <name> instead of nobody when dropping privileges\n");
  printf("--disable-cache                     | Disable flow cache for avoid merging flows. This option\n"
	 "                                    | is available only in collector/proxy mode\n"
	 "                                    | (i.e. use -i none)\n");

  printf("\nFurther plugin available command line options\n");
  printf("---------------------------------------------------\n");
  initPlugins(0, NULL);
  dumpPluginHelp();

  printf("\n\nNote on interface indexes and (router) MAC/IP addresses\n"
	 "---------------------------------------------------\n"
	 "Flags -u and -Q are used to specify the SNMP interface identifiers for emitted flows.\n"
	 "However using --if-networks it is possible to specify an interface identifier to which\n"
	 "a MAC address or IP network is bound. The syntax of --if-networks is:\n"
	 " <MAC|IP/mask>@<interfaceId> where multiple entries can be separated by a comma (,).\n"
	 "Example: --if-networks \"AA:BB:CC:DD:EE:FF@3,192.168.0.0/24@2\" or\n"
	 "--if-networks @<filename> where <filename> is a file path containing the networks\n"
	 "specified using the above format.\n");

#ifdef WIN32
  (void)printAvailableInterfaces("-1");
#endif
  printf("\nNetFlow v9/IPFIX format [-T]"
	 "\n----------------"
	 "\nThe following options can be used to specify the format:\n"
	 "\n ID   NetFlow Label               IPFIX Label                   Description\n"
	 "-------------------------------------------------------------------------------\n");

  printTemplateInfo(ver9_templates, 0);
  dumpPluginTemplates();

  printf("\nMajor protocol (%%L7_PROTO) symbolic mapping 0...%d:\n", IPOQUE_MAX_SUPPORTED_PROTOCOLS-1);

  for(i=0; i<IPOQUE_MAX_SUPPORTED_PROTOCOLS; i++)
    printf("[%3d] %s\n", i, prot_long_str[i]);

  /* ************************************************ */

  printf("\nExample: nprobe -T \"%s\"\n", DEFAULT_V9_IPV4_TEMPLATE);

  printf("\n");
  printf("nProbe shut down\n");

  exit(0);
}

/* ****************************************************** */

static void dumpStats(char *path, u_int32_t tot_pkts, u_int32_t tot_bytes) {
  FILE *fd = fopen(path, "w");

  if(fd) {
    fprintf(fd,
	    "# time totalPkts totalBytes totFlows\n"
	    "%u %u %u %u\n",
	    (unsigned int)time(NULL),
	    tot_pkts, tot_bytes,
	    (unsigned int)readWriteGlobals->totFlows);
    fclose(fd);
  } else
    traceEvent(TRACE_WARNING, "Unable to create file %s", path);
}

/* ****************************************************** */

static printProcessingStats(void) {
  u_int32_t tot_pkts = 0, tot_bytes = 0;
  u_int num_collected_pkts = 0, i;

  for(i=0; i<readOnlyGlobals.numProcessThreads; i++) {
    tot_pkts  += readWriteGlobals->accumulateStats[i].pkts;
    tot_bytes += readWriteGlobals->accumulateStats[i].bytes;
  }

  for(i=0; i<readOnlyGlobals.numProcessThreads; i++)
    num_collected_pkts += readWriteGlobals->collectedPkts[i];

  traceEvent(TRACE_NORMAL, "Processed packets: %u (max bucket search: %d)",
	     (unsigned long)tot_pkts, readWriteGlobals->maxBucketSearch);

  traceEvent(TRACE_NORMAL, "Flow export stats: [%u bytes][%u pkts][%u flows]",
	     (unsigned long)readWriteGlobals->flowExportStats.totExportedFlowBytes,
	     (unsigned long)readWriteGlobals->flowExportStats.totExportedFlowPkts,
	     (unsigned long)readWriteGlobals->flowExportStats.totExportedFlows);

  if(readOnlyGlobals.collectorInPort > 0)
    traceEvent(TRACE_NORMAL, "Flow collection: [collected pkts: %u][processed flows: %u]",
	       num_collected_pkts, readWriteGlobals->collectionStats.num_flows_processed);

  traceEvent(TRACE_NORMAL, "Flow drop stats:   [%u bytes][%u pkts][%u flows]",
	     (unsigned long)readWriteGlobals->probeStats.totFlowBytesDropped,
	     (unsigned long)readWriteGlobals->probeStats.totFlowPktsDropped,
	     (unsigned long)readWriteGlobals->probeStats.totFlowDropped);

  traceEvent(TRACE_NORMAL, "Total flow stats:  [%u bytes][%u pkts][%u flows]",
	     (unsigned long)readWriteGlobals->probeStats.totFlowBytesDropped +
	     (unsigned long)readWriteGlobals->flowExportStats.totExportedFlowBytes,
	     (unsigned long)readWriteGlobals->probeStats.totFlowPktsDropped +
	     (unsigned long)readWriteGlobals->flowExportStats.totExportedFlowPkts,
	     (unsigned long)readWriteGlobals->probeStats.totFlowDropped +
	     (unsigned long)readWriteGlobals->flowExportStats.totExportedFlows);

  if(readOnlyGlobals.tracePerformance && (tot_pkts > 0)) {
    static unsigned long last_pkts = 0;
    ticks tot;

    pthread_rwlock_wrlock(&readOnlyGlobals.ticksLock);

    tot = readOnlyGlobals.decodeTicks
      + readOnlyGlobals.processingWithFlowCreationTicks
      + readOnlyGlobals.processingWoFlowCreationTicks
      ;

    if(tot > 0) {
      if(last_pkts == 0) last_pkts = tot_pkts;
      last_pkts = tot_pkts - last_pkts;

      if(last_pkts > 0) {
	traceEvent(TRACE_NORMAL, "---------------------------------");
	traceEvent(TRACE_NORMAL, "Decode ticks:     %.2f ticks/pkt [%.2f %%]",
		   (float)readOnlyGlobals.decodeTicks / (float)last_pkts,
		   (float)(readOnlyGlobals.decodeTicks*100)/(float)tot);

	if(readOnlyGlobals.num_pkts_without_flow_creation == 0) readOnlyGlobals.num_pkts_without_flow_creation = 1;
	traceEvent(TRACE_NORMAL, "Pkt Processing w/o Flow Creation: %.2f ticks/pkt [%.2f %%]",
		   (float)readOnlyGlobals.processingWoFlowCreationTicks / (float)readOnlyGlobals.num_pkts_without_flow_creation,
		   (float)(readOnlyGlobals.processingWoFlowCreationTicks*100) / (float)tot);

	if(readOnlyGlobals.num_pkts_with_flow_creation == 0) readOnlyGlobals.num_pkts_with_flow_creation = 1;
	traceEvent(TRACE_NORMAL, "Pkt Processing with Flow Creation: %.2f ticks/pkt [%.2f %%]",
		   (float)readOnlyGlobals.processingWithFlowCreationTicks / (float)readOnlyGlobals.num_pkts_with_flow_creation,
		   (float)(readOnlyGlobals.processingWithFlowCreationTicks*100) / (float)tot);

	if(readOnlyGlobals.num_allocated_buckets == 0) readOnlyGlobals.num_allocated_buckets = 1;
	traceEvent(TRACE_NORMAL, "Bucket Allocation: %.2f ticks/bkt",
		   (float)readOnlyGlobals.bucketAllocationTicks / (float)readOnlyGlobals.num_allocated_buckets);

	if(readOnlyGlobals.num_malloced_buckets == 0) readOnlyGlobals.num_malloced_buckets = 1;
	traceEvent(TRACE_NORMAL, "Bucket Malloc: %.2f ticks/bkt",
		   (float)readOnlyGlobals.bucketMallocTicks / (float)readOnlyGlobals.num_malloced_buckets);

	if(readOnlyGlobals.num_exported_buckets == 0) readOnlyGlobals.num_exported_buckets = 1;
	traceEvent(TRACE_NORMAL, "Bucket Export: %.2f ticks/bkt",
		   (float)readOnlyGlobals.bucketExportTicks / (float)readOnlyGlobals.num_exported_buckets);

	if(readOnlyGlobals.num_purged_buckets == 0) readOnlyGlobals.num_purged_buckets = 1;
	traceEvent(TRACE_NORMAL, "Bucket Purge: %.2f ticks/bkt",
		   (float)readOnlyGlobals.bucketPurgeTicks / (float)readOnlyGlobals.num_purged_buckets);

	traceEvent(TRACE_NORMAL, "Total ticks:      %.2f ticks/pkt",
		   (float)tot / (float)last_pkts);
	traceEvent(TRACE_NORMAL, "---------------------------------");
      }

    }

    last_pkts = tot_pkts, readOnlyGlobals.decodeTicks = readOnlyGlobals.processingWithFlowCreationTicks
      = readOnlyGlobals.processingWoFlowCreationTicks =
      readOnlyGlobals.bucketAllocationTicks = 
      readOnlyGlobals.bucketMallocTicks =
      readOnlyGlobals.bucketPurgeTicks  = readOnlyGlobals.bucketExportTicks = 0;
    readOnlyGlobals.num_exported_buckets = readOnlyGlobals.num_purged_buckets =
      readOnlyGlobals.num_allocated_buckets =
      readOnlyGlobals.num_malloced_buckets =
      readOnlyGlobals.num_pkts_without_flow_creation = readOnlyGlobals.num_pkts_with_flow_creation = 0;

    pthread_rwlock_unlock(&readOnlyGlobals.ticksLock);
  }

  printCaptureStats();
}

/* ****************************************************** */

static void printStats(int force) {
  time_t now = time(NULL), nowDiff;
  char pktBuf[32], buf[1024] = { 0 };
  u_int tot_frags, i;
  Counter tot_pkts = 0, tot_bytes = 0, current_pkts = 0, current_bytes = 0;

  readWriteGlobals->now = now;
  nowDiff = now-readOnlyGlobals.initialSniffTime.tv_sec;

  if(!force) {
    /* Wait at least 5 seconds */
    if((nowDiff < 5) /* || (readWriteGlobals->accumulateStats.pkts == 0) */)
      return;
  }

  if(readOnlyGlobals.traceMode) {
    if(readOnlyGlobals.numProcessThreads > 1)
      traceEvent(TRACE_NORMAL, "---------------------------------");
  }

  for(i=0; i<readOnlyGlobals.numProcessThreads; i++) {
    tot_pkts      += readWriteGlobals->accumulateStats[i].pkts;
    tot_bytes     += readWriteGlobals->accumulateStats[i].bytes;
    current_pkts  += readWriteGlobals->currentPkts[i];
    current_bytes += readWriteGlobals->currentBytes[i];

    readWriteGlobals->currentPkts[i] = 0, readWriteGlobals->currentBytes[i] = 0;

    if(readOnlyGlobals.numProcessThreads > 1)
      traceEvent(TRACE_NORMAL, "Average traffic: [queue %u][%s pps][%s/sec]",
		 i,
		 formatPackets((float)readWriteGlobals->accumulateStats[i].pkts/nowDiff, pktBuf),
		 formatTraffic((float)(8*readWriteGlobals->accumulateStats[i].bytes)/(float)nowDiff, 1, buf));
  }

  if(readOnlyGlobals.dump_stats_path != NULL)
    dumpStats(readOnlyGlobals.dump_stats_path, tot_pkts, tot_bytes);

  if(readOnlyGlobals.traceMode) {
    u_int bucketsAllocated;

    if(readOnlyGlobals.numProcessThreads == 1) traceEvent(TRACE_NORMAL, "---------------------------------");
    traceEvent(TRACE_NORMAL, "Average traffic: [%s pps][%s/sec]",
	       formatPackets((float)(tot_pkts/nowDiff), pktBuf),
	       formatTraffic((float)(8*tot_bytes)/(float)nowDiff, 1, buf));

    nowDiff = now-readWriteGlobals->lastSample;
    if(nowDiff == 0) nowDiff = 1;
    traceEvent(TRACE_NORMAL, "Current traffic: [%s pps][%s/sec]",
	       formatPackets((float)(current_pkts/nowDiff), pktBuf),
	       formatTraffic((float)(8*current_bytes)/(float)nowDiff, 1, buf));
    readWriteGlobals->lastSample = readWriteGlobals->now;

    traceEvent(TRACE_NORMAL, "Current flow export rate: [%.1f flows/sec]",
	       (float)readWriteGlobals->totFlowsRate/nowDiff);

    traceEvent(TRACE_NORMAL, "Flow drops: [export queue too long=%u][too many flows=%u]",
	       readWriteGlobals->probeStats.totFlowDropped,
	       readWriteGlobals->probeStats.droppedPktsTooManyFlows);
    readWriteGlobals->totFlowsRate = 0;
    traceEvent(TRACE_NORMAL, "Export Queue: %u/%d [%.1f %%]",
	       readWriteGlobals->exportBucketsLen,
	       readOnlyGlobals.maxExportQueueLen,
	       ((float)(readWriteGlobals->exportBucketsLen * 100))/(float)readOnlyGlobals.maxExportQueueLen);

    for(tot_frags=0, i=0; i<NUM_FRAGMENT_LISTS; i++) tot_frags += readWriteGlobals->fragmentListLen[i];
    for(bucketsAllocated=0, i=0; i<readOnlyGlobals.numProcessThreads; i++)
      bucketsAllocated += readWriteGlobals->bucketsAllocated[i];

    traceEvent(TRACE_NORMAL, "Flow Buckets: [active=%u][allocated=%u][toBeExported=%u][frags=%u]",
	       bucketsAllocated-readWriteGlobals->exportBucketsLen,
	       bucketsAllocated, readWriteGlobals->exportBucketsLen, tot_frags);

    dumpPluginStats();

    buf[0] = '\0';
    for(i=0; i<readOnlyGlobals.numProcessThreads; i++) {
      if(readWriteGlobals->collectedPkts[i] > 0) {
	u_int len = strlen(buf);

	snprintf(&buf[len], sizeof(buf)-len, "%lu@%d ",
		 readWriteGlobals->collectedPkts[i], i);
      }
    }

    if(buf[0] != '\0')
      traceEvent(TRACE_NORMAL, "Collector Threads: [ %s]", buf);
  }

  if(readOnlyGlobals.traceMode) {
    printProcessingStats();
  } else {
    if(readWriteGlobals->maxBucketSearch > readWriteGlobals->lastMaxBucketSearch) {
      traceEvent(TRACE_INFO, "Max bucket search: %d slots (for better "
		 "performance a larger value for -w)",
		 readWriteGlobals->maxBucketSearch);
      readWriteGlobals->lastMaxBucketSearch = readWriteGlobals->maxBucketSearch;
    }
  }

  readWriteGlobals->maxBucketSearch = 0; /* reset */

  if(readOnlyGlobals.pcapPtr && (!readOnlyGlobals.traceMode))
    printPcapStats(readOnlyGlobals.pcapPtr);
}

/* ****************************************************** */

int resolveIpV4Address(char *addr, int port) {
  struct hostent *hostAddr;
  struct in_addr dstAddr;

  if((hostAddr = gethostbyname(addr)) == NULL) {
    traceEvent(TRACE_ERROR, "Unable to resolve address '%s'\n", addr);
    return(-1);
  }

  memset(&readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors], 0, sizeof(CollectorAddress));
  memcpy(&dstAddr.s_addr, hostAddr->h_addr_list[0], hostAddr->h_length);
  readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd = -1;
  readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].isIPv6 = 0;
  readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].u.v4Address.sin_addr.s_addr = dstAddr.s_addr;
  readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].u.v4Address.sin_family      = AF_INET;
  readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].u.v4Address.sin_port        = (int)htons(port);

  return(0);
}

/* ****************************************************** */

int resolveIpV6Address(char *addr, int port, int *isIpV6Address) {
  int errnum;
  struct addrinfo hints, *res;

  if((readOnlyGlobals.useIpV6 == 0) || !strstr(addr, ":")) {
    (*isIpV6Address) = 0;
    return(resolveIpV4Address(addr, port));
  }

  (*isIpV6Address) = 0;
  memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_UNSPEC;
  hints.ai_socktype = SOCK_DGRAM;

  errnum = getaddrinfo(addr, NULL, &hints, &res);
  if(errnum != 0) {
    traceEvent(TRACE_INFO, "Unable to resolve address '%s' [error=%d]\n",
	       addr, errnum);
    return(-1);
  }

  if(res->ai_family == PF_INET6) {
    (*isIpV6Address) = 1;
    memset(&readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors], 0, sizeof(CollectorAddress));
    readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].isIPv6 = 1;
    memcpy(&readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].u.v6Address, res->ai_addr, res->ai_addrlen);
    freeaddrinfo(res);
    readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].u.v6Address.sin6_port = (int)htons(port);
    return(0);
  } else {
    freeaddrinfo(res);
    (*isIpV6Address) = 0;
    return(resolveIpV4Address(addr, port));
  }
}

/* ****************************************************** */

#define PROTO_UDP_URL       "udp://"
#define PROTO_TCP_URL       "tcp://"
#define PROTO_SCTP_URL      "sctp://"

int initNetFlow(char* addr, int port) {
  int sockopt = 1, rc, isIpV6Address = 0;
  char *address;
  u_char transport = TRANSPORT_UDP;

  if(readOnlyGlobals.numCollectors >= MAX_NUM_COLLECTORS) {
    traceEvent(TRACE_INFO,
	       "Unable to define further collector address "
	       "(max %d collectors allowed)\n", MAX_NUM_COLLECTORS);
    return(-1);
  }

  if(strncmp(addr, PROTO_UDP_URL, strlen(PROTO_UDP_URL)) == 0)
    transport = TRANSPORT_UDP, address = &addr[strlen(PROTO_UDP_URL)];
  else if(strncmp(addr, PROTO_TCP_URL, strlen(PROTO_TCP_URL)) == 0)
    transport = TRANSPORT_TCP, address = &addr[strlen(PROTO_TCP_URL)];
  else if(strncmp(addr, PROTO_SCTP_URL, strlen(PROTO_SCTP_URL)) == 0) {
#ifdef HAVE_SCTP
    transport = TRANSPORT_SCTP;
#else
    traceEvent(TRACE_ERROR, "SCTP isn't supported on your system. Using UDP.");
    transport = TRANSPORT_UDP;
#endif
    address = &addr[strlen(PROTO_SCTP_URL)];
  } else
    transport = TRANSPORT_UDP, address = addr;

  if(readOnlyGlobals.useIpV6) {
    rc = resolveIpV6Address(address, port, &isIpV6Address);
    if(!isIpV6Address) readOnlyGlobals.useIpV6 = 0;
  } else
    rc = resolveIpV4Address(address, port);

  if(rc != 0)  return(-1);

  /* Initialize the socket descriptor, so that it looks like it is not opened yet */
  readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd = -1;

  if((readOnlyGlobals.sockIn.sin_addr.s_addr == 0)
#ifdef IP_HDRINCL
     /* Check only if packet reforging is enabled */
     && (!readOnlyGlobals.nfLitePluginEnabled)
#endif
     ) {
    if(readOnlyGlobals.useIpV6) {
      if(transport == TRANSPORT_UDP)
	readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd = socket(AF_INET6, SOCK_DGRAM, 0);
      else if(transport == TRANSPORT_TCP)
	readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd = socket(AF_INET6, SOCK_STREAM, 0);
#ifdef HAVE_SCTP
      else if(transport == TRANSPORT_SCTP)
	readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd = socket(AF_INET6, SOCK_SEQPACKET,
										   IPPROTO_SCTP);
#endif
    }
    if(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd == -1) {
      readOnlyGlobals.useIpV6 = 0; /* No IPv6 ? */
      if(transport == TRANSPORT_UDP)
	readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd = socket(AF_INET, SOCK_DGRAM, 0);
      else if(transport == TRANSPORT_TCP)
	readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd = socket(AF_INET, SOCK_STREAM, 0);
#ifdef HAVE_SCTP
      else if(transport == TRANSPORT_SCTP)
	readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd = socket(AF_INET, SOCK_SEQPACKET,
										   IPPROTO_SCTP);
#endif
    }
  }
#ifdef IP_HDRINCL
  else {
    int tmp = 1;

    if(transport != TRANSPORT_UDP) {
      transport = TRANSPORT_UDP;
      traceEvent(TRACE_WARNING,
		 "Unable to use a transport different from UDP");
      traceEvent(TRACE_WARNING, "when -q is used. Reverting to UDP.");
    }

    readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd = socket(PF_INET, SOCK_RAW, IPPROTO_UDP);

    if(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd == -1) {
      traceEvent(TRACE_ERROR, "Fatal error while creating socket (%s).",
		 strerror(errno));
#ifndef WIN32
      if((getuid() && geteuid()) || setuid (0)) {
	traceEvent(TRACE_ERROR, "You probably need superuser capabilities. "
		   "Please try again.");
      }
#endif

      exit(-1);
    }

    transport = TRANSPORT_UDP_RAW;
    /* Tell that we specify the IP header */
    setsockopt(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd, 0, IP_HDRINCL,
	       &tmp, sizeof(tmp));
  }
#endif

  setsockopt(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd, SOL_SOCKET, SO_REUSEADDR,
	     (char *)&sockopt, sizeof(sockopt));

  if(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd == -1) {
    traceEvent(TRACE_INFO, "Fatal error while creating socket (%s).",
	       strerror(errno));
    exit(-1);
  }

  readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].transport = transport;

  if(transport == TRANSPORT_TCP) {
    int rc;

    traceEvent(TRACE_INFO, "Connecting to %s:%d...", addr, port);

    if(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].isIPv6) {
      char col[100];

      inet_ntop(AF_INET6, &readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].u.v6Address, col, sizeof(col));
      rc = connect(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd,
		   (struct sockaddr *)&readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].u.v6Address,
		   sizeof(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].u.v6Address));
    } else
      rc = connect(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd,
		   (struct sockaddr *)&readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].u.v4Address,
		   sizeof(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].u.v4Address));

    if(rc == -1) {
      char msg[256], buf[64];

      snprintf(msg, sizeof(msg), "Connection failed with remote peer %s [%s]. Leaving.",
	       CollectorAddress2Str(&readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors], buf, sizeof(buf)),
	       strerror(errno));

      traceEvent(TRACE_ERROR, "%s", msg);
      dumpLogEvent(collector_connection_error, severity_error, msg);
      close(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd);
      exit(-1);
    } else {
      char buf[64], msg[256];

      snprintf(msg, sizeof(msg), "Succesfully connected with remote collector %s",
	       CollectorAddress2Str(&readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors], buf, sizeof(buf)));

      dumpLogEvent(collector_connected, severity_info, msg);
    }
  } else if(transport == TRANSPORT_UDP) {
    maximize_socket_buffer(readOnlyGlobals.netFlowDest[readOnlyGlobals.numCollectors].sockFd, SO_SNDBUF);
  }

  readOnlyGlobals.numCollectors++;

  if(strstr(address, ":"))
    traceEvent(TRACE_INFO, "Exporting flows towards [%s]:%d using %s",
	       addr, port,
	       ((transport == TRANSPORT_UDP)
#ifdef IP_HDRINCL
		|| (transport == TRANSPORT_UDP_RAW)
#endif
		) ? "UDP" :
	       (transport == TRANSPORT_TCP ? "TCP" : "SCTP"));
  else
    traceEvent(TRACE_INFO, "Exporting flows towards %s:%d using %s",
	       addr, port,
	       ((transport == TRANSPORT_UDP)
#ifdef IP_HDRINCL
		|| (transport == TRANSPORT_UDP_RAW)
#endif
		) ? "UDP" :
	       (transport == TRANSPORT_TCP ? "TCP" : "SCTP"));

  return(0);
}

/* ****************************************************** */

void printHash(int idx) {
  u_int i;

  for(i = 0; i<readOnlyGlobals.flowHashSize; i++) {
    if(readWriteGlobals->theFlowHash[idx][i] != NULL)
      printf("readWriteGlobals->theFlowHash[%4d]\n", i);
  }
}

/* ****************************************************** */

void dumpBuffer(char *buffer, int bufferLength) {
  int i;

  if(bufferLength > 512) bufferLength = 512;

  for(i=0; i<bufferLength; i++) {
    if(!(i % 8)) printf("\n");
    printf("%3d[%02x] ", i, buffer[i] & 0xFF );
  }

  printf("\n");
}

/* ****************************************************** */

static void readPcapFileList(const char * filename) {
  char line[512];

  FILE *fd = fopen(filename, "r");

  if(fd != NULL) {
    struct fileList *fl, *prev;

    while(!feof(fd)) {
      int i, bad_line;

      if(fgets(line, sizeof(line)-1, fd) == NULL) continue;
      if((line[0] == '#') || (line[0] == '\n')) continue;

      bad_line = 0;

      for(i=0; i<strlen(line); i++) {
	if(!isascii(line[i])) {
	  bad_line = 1;
	  break;
	}
      }

      if(bad_line) {
	traceEvent(TRACE_ERROR, "Your --pcap-file-list %s contains binary data: discarded", filename);
	fclose(fd);
	return;
      }

      while(strlen(line) && (line[strlen(line)-1] == '\n')) line[strlen(line)-1] = '\0';

      fl = (struct fileList*)malloc(sizeof(struct fileList));

      if(!fl) {
	traceEvent(TRACE_ERROR, "Not enough memory parsing --pcap-file-list argument");
	fclose(fd);
	return;
      }

      fl->path = strdup(line);

      if(!fl->path) {
	free(fl);
        traceEvent(TRACE_ERROR, "Not enough memory parsing --pcap-file-list argument");
        fclose(fd);
        return;
      }

      fl->next = NULL;

      if(readOnlyGlobals.pcapFileList) {
	prev = readOnlyGlobals.pcapFileList;
	while(prev != NULL) {
	  if(prev->next)
	    prev = prev->next;
	  else
	    break;
	}

	prev->next = fl;
      } else
	readOnlyGlobals.pcapFileList = fl;
    }

    fclose(fd);
  } else
    traceEvent(TRACE_ERROR, "Unable to open file %s", optarg);
}

/* ****************************************************** */

static void initDefaults(void) {
  memset(&readOnlyGlobals.sockIn, 0, sizeof(readOnlyGlobals.sockIn));

  /* Set defaults */
  readOnlyGlobals.netFlowVersion = 5; /* NetFlow v5 */
  readOnlyGlobals.bidirectionalFlows = 0;
  readOnlyGlobals.ignorePorts = readOnlyGlobals.ignorePorts = readOnlyGlobals.ignoreProtocol = 0;
  readOnlyGlobals.ignoreIP = readOnlyGlobals.ignoreIP = 0, readOnlyGlobals.ignoreTos = 0;
#ifdef HAVE_GEOIP
  readOnlyGlobals.geo_ip_asn_db = NULL;
#endif
  readOnlyGlobals.numCollectors = 0;
  readOnlyGlobals.flowHashSize = DEFAULT_HASH_SIZE;
  readOnlyGlobals.hostHashSize = readOnlyGlobals.flowHashSize/2;
  readOnlyGlobals.initialSniffTime.tv_sec = 0; /* Set it with the first incoming packet */
  readOnlyGlobals.snaplen = PCAP_DEFAULT_SNAPLEN;
  readOnlyGlobals.pcapFileList = NULL;
  readOnlyGlobals.pcapFile = NULL;
  readOnlyGlobals.reflectorMode = 0;
  readOnlyGlobals.minFlowSize = 0;
  readOnlyGlobals.traceMode = 0;
  readOnlyGlobals.flowExportDelay = 1, readOnlyGlobals.packetFlowGroup = 1;
  readOnlyGlobals.engineType = 0, readOnlyGlobals.engineId = time(NULL) /* dynamic */;
  readOnlyGlobals.useNetFlow = 0xFF;
  readOnlyGlobals.use_vlanId_as_ifId = vlan_disabled;
  readOnlyGlobals.stringTemplateV4 = readOnlyGlobals.stringTemplateV6 = NULL;
  readOnlyGlobals.dirPath = NULL;
  readOnlyGlobals.minNumFlowsPerPacket = -1;
  readOnlyGlobals.pktSampleRate = 1;
  readOnlyGlobals.fakePktSampling = 0;
  readOnlyGlobals.flowSampleRate = 1;
  readOnlyGlobals.numInterfaceNetworks = 0;
  readOnlyGlobals.numBlacklistNetworks = 0;
  readOnlyGlobals.roundPacketLenWithIPHeaderLen = 1;
  readOnlyGlobals.maxExportQueueLen = MAX_EXPORT_QUEUE_LEN;
  readOnlyGlobals.unprivilegedUser = strdup("nobody");

  readWriteGlobals->numFlows = 0;
  readWriteGlobals->lastExportTime.tv_sec = 0, readWriteGlobals->lastExportTime.tv_usec = 0;
  readWriteGlobals->num_src_mac_export = 0;

#ifdef HAVE_PF_RING
  readOnlyGlobals.cluster_id = -1;
#endif

  initAS();
}

/* ****************************************************** */

static int parseOptions(int argc, char* argv[], u_int8_t reparse_options) {
  int id;
  char *theItem;
  char line[2048];
  FILE *fd;
  int opt, i, opt_n = 0;
  u_char mandatoryParamOk=0;

  if(!reparse_options)
    initDefaults();

  optind = 0;
#ifdef HAVE_OPTRESET
  optreset = 1; /* Make sure getopt read options again */
#endif

  readOnlyGlobals.argc = 0;
  readOnlyGlobals.argv = (char**)malloc(sizeof(char*)*MAX_NUM_OPTIONS);
  memset(readOnlyGlobals.argv, 0, sizeof(char*)*MAX_NUM_OPTIONS);

  if(readOnlyGlobals.argv == NULL) return(-1);

  if((argc == 2) && (argv[1][0] != '-')) {
    char *tok, cont=1;

    fd = fopen(argv[1], "r");

    if(fd == NULL) {
      traceEvent(TRACE_ERROR, "Unable to read config. file %s", argv[1]);
      exit(-1);
    }

    readOnlyGlobals.argv[readOnlyGlobals.argc++] = strdup("nprobe");

    while(cont && fgets(line, sizeof(line), fd)) {
      /* printf("line='%s'\n", line); */

      /*
	Config files accept both
	<option>=<value>
	and
	<option> <value>
      */
      i = 0;
      while(line[i] != '\0') {
	if(line[i] == '=')
	  break;
	else if(line[i] == ' ') {
	  line[i] = '=';
	  break;
	}

	i++;
      }

      tok = strtok(line, "=");

      while(tok != NULL) {
	int len;
	char *argument;

	if(readOnlyGlobals.argc >= MAX_NUM_OPTIONS) {
	  int i;

	  traceEvent(TRACE_ERROR, "Command line too long [%u arguments]", readOnlyGlobals.argc);

	  for(i=0; i<readOnlyGlobals.argc; i++)
	    traceEvent(TRACE_ERROR, "[%d][%s]", i, readOnlyGlobals.argv[i]);
	  cont = 0; break;
	}

	len = strlen(tok)-1;
	if(tok[len] == '\n') tok[len] = '\0';

	if((tok[0] == '\"') && (tok[strlen(tok)-1] == '\"')) {
	  tok[strlen(tok)-1] = '\0';
	  argument = &tok[1];
	} else
	  argument = tok;

	if(argument && (argument[0] != '\0')) {
	  /* traceEvent(TRACE_NORMAL, "readOnlyGlobals.argv[%d]='%s'", readOnlyGlobals.argc, argument); */
	  readOnlyGlobals.argv[readOnlyGlobals.argc++] = strdup(argument);
	}

	tok = strtok(NULL, "\n");
      }
    }

    fclose(fd);
  } else {
    if(reparse_options) {
      traceEvent(TRACE_WARNING, "Command line options can be reloaded only when");
      traceEvent(TRACE_WARNING, "the probe is started from a configuration file");
      traceEvent(TRACE_WARNING, "Please use nprobe <configuration file>");
      return(-1);
    }

    if(argc >= MAX_NUM_OPTIONS)
      readOnlyGlobals.argc = MAX_NUM_OPTIONS-1;
    else
      readOnlyGlobals.argc = argc;

    /* Copy arguments */
    for(i=0; i<readOnlyGlobals.argc; i++) {
      readOnlyGlobals.argv[i] = strdup(argv[i]);
    }
  }

  readOnlyGlobals.useIpV6 = 1;
  optarg = NULL;

  while((opt = getopt_long(readOnlyGlobals.argc, readOnlyGlobals.argv,
			   "A:ab:B:c"
			   "C:d:D:e:E:f:F:g:hi:I:l:L:"
#ifdef IP_HDRINCL
			   "q:"
#endif
			   "M:m:Nn:o:O:p:P:Q:rR:s:S:t:T:u:U:w:Wx:vV:z:"
#ifndef WIN32
			   "G"
#endif
#if defined(linux) || defined(__linux__)
			   "X4:"
#endif

			   "1:2:3:a:"
			   "5678:9:!:@"
#if !defined(WIN32)
			   "$:\\"
#endif
			   ,
			   long_options,
			   NULL
			   )) != EOF) {
    if(reparse_options) {
      u_int discard_option;

      switch(opt) {
      case 'b':
      case 'B':
      case 'd':
      case 'e':
      case 'E':
      case 'F':
      case 'l':
      case 'm':
      case 'M':
      case 'o':
      case 'Q':
      case 's':
      case 'S':
      case 'T':
      case 't':
      case 'u':
      case 'U':
      case '5':
      case '7':
      case '8':
      case '^':
      case '{':
      case '}':
      case '+':
	discard_option = 0;
	break;

      default:
	discard_option = 1;
      }

      if(discard_option) {
	traceEvent(TRACE_WARNING, "The %c option cannot be modified at runtime: ignored", opt);
	continue;
      }
    }

    switch(opt) {
    case '1':
      parseInterfaceAddressLists(optarg);
      break;
    case '2':
      readOnlyGlobals.capture_num_packet_and_quit = atoi(optarg);
      break;
    case '3':
      readOnlyGlobals.collectorInPort = atoi(optarg);
      break;
#ifdef linux
    case '4':
      readOnlyGlobals.cpuAffinity = strdup(optarg);
      break;
#endif
    case '5':
      readOnlyGlobals.tunnel_mode = 1;
      break;
    case '6':
      readOnlyGlobals.promisc_mode = 0;
      break;
    case '7':
      readOnlyGlobals.smart_udp_frags_mode = 1;
      break;
    case '8':
      readOnlyGlobals.ipsec_auth_data_len = atoi(optarg);
      break;
    case '9':
      readOnlyGlobals.dump_stats_path = strdup(optarg);
      break;
    case '!':
      parseBlacklistNetworks(optarg);
      break;
    case '@':
      if(!strcmp(optarg, "inner"))
	readOnlyGlobals.use_vlanId_as_ifId = inner_vlan;
      else
	readOnlyGlobals.use_vlanId_as_ifId = outer_vlan;
      break;
    case '&':
      readOnlyGlobals.l7.discard_unknown_flows = atoi(optarg);
      if(readOnlyGlobals.l7.discard_unknown_flows > 2) {
	traceEvent(TRACE_WARNING, "The mode for --discard-unknown-flows is out of range: ignored.");
	readOnlyGlobals.l7.discard_unknown_flows = 0;
      }
      break;
    case '$':
      readPcapFileList(optarg);
      break;
    case '\\':
      readOnlyGlobals.do_not_drop_privileges = 1;
      break;
    case '^':
      if(readOnlyGlobals.csv_separator) free(readOnlyGlobals.csv_separator);
      readOnlyGlobals.csv_separator = strdup(optarg);
      break;
    case ',':
      readCities(optarg);
      break;
#ifdef HAVE_FASTBIT
    case '[':
      {
	int len;

	if(readOnlyGlobals.fastbit_dump_directory != NULL)
	  free(readOnlyGlobals.fastbit_dump_directory);

	readOnlyGlobals.fastbit_dump_directory = strdup(optarg);
	len = strlen(readOnlyGlobals.fastbit_dump_directory);
	if(readOnlyGlobals.fastbit_dump_directory[len-1] ==
#ifdef WIN32
	   '\\'
#else
	   '/'
#endif
	   )
	  readOnlyGlobals.fastbit_dump_directory[len-1] = '\0';
      }
      break;

    case ']':
      readOnlyGlobals.fastbit_mins_rotation = atoi(optarg);
      if(readOnlyGlobals.fastbit_mins_rotation == 0)
	readOnlyGlobals.fastbit_mins_rotation = 1;
      break;

    case '(':
      readOnlyGlobals.fastbit_dump_template = strdup(optarg);
      break;

#ifndef WIN32
    case ')':
      readOnlyGlobals.fastbit_index_directory = 1;
      readOnlyGlobals.fastbit_index_columns = strdup(optarg);
      break;
#endif

    case '#':
      readOnlyGlobals.fastbit_exec = strdup(optarg);
      break;
#endif

    case '{':
      readOnlyGlobals.bidirectionalFlows = 1;
      break;
    case '}':
      readOnlyGlobals.accountL2Traffic = 1;
      break;
    case '=':
      {
	FILE *fd = fopen(optarg, "w");

	if(fd) {
	  initPlugins(0, NULL);
	  printMetadata(fd);
	  fclose(fd);
	  traceEvent(TRACE_NORMAL, "Dumped metadata on file %s", optarg);
	} else
	  traceEvent(TRACE_ERROR, "Unable to create file %s", optarg);

	exit(0);
      }
      break;

    case '+':
      {
	char *old = readOnlyGlobals.eventLogPath;

	readOnlyGlobals.eventLogPath = strdup(optarg);
	if(old == NULL) free(old);
      }
      break;

    case 'A':
      readASs(optarg);
      break;
    case 'a':
      readOnlyGlobals.reflectorMode = 1;
      break;
    case 'b':
      i = atoi(optarg);
      if(i > 2) i = 2;
      switch(i) {
      case 1:
	readOnlyGlobals.traceMode = 1, readOnlyGlobals.traceLevel = 5;
	break;
      case 2:
	readOnlyGlobals.traceMode = 2, readOnlyGlobals.traceLevel = 5;
	break;
      case 0:
      default:
	readOnlyGlobals.traceMode = 0, readOnlyGlobals.traceLevel = 2;
	break;
      }
      break;
    case 'B':
      readOnlyGlobals.packetFlowGroup = atoi(optarg);
      break;
    case 'c':
      if(readOnlyGlobals.numLocalNetworks == 0) {
	traceEvent(TRACE_WARNING, "Ignored -c: it must be specified after -L");
      } else
	readOnlyGlobals.setAllNonLocalHostsToZero = 1;
      break;
    case 'C':
      readOnlyGlobals.flowLockFile = strdup(optarg);
      break;
    case 'P':
      if(optarg[0] != '\0') {
	readOnlyGlobals.dirPath = strdup(optarg);
	if(readOnlyGlobals.dirPath[strlen(readOnlyGlobals.dirPath)-1] == '/') readOnlyGlobals.dirPath[strlen(readOnlyGlobals.dirPath)-1] = '\0';
      }
      break;
    case 'D':
      if(optarg[0] == 't')      readOnlyGlobals.dumpFormat = text_format;
      else if(optarg[0] == 'd') readOnlyGlobals.dumpFormat = sqlite_format;
      else if(optarg[0] == 'b') readOnlyGlobals.dumpFormat = binary_format;
      else if(optarg[0] == 'B') readOnlyGlobals.dumpFormat = binary_core_flow_format;
      else traceEvent(TRACE_WARNING, "Invalid -D option '%s': ignored", optarg);
      break;
    case 'd':
      readOnlyGlobals.idleTimeout = atoi(optarg);
      break;
    case 'E':
      theItem = strtok(optarg, ":");
      if(theItem == NULL) {
	traceEvent(TRACE_WARNING, "Wrong engine specified (-E flag): see help.");
      } else {
	readOnlyGlobals.engineType = (u_int8_t)atoi(theItem);
	theItem = strtok(NULL, ":");

	if(theItem == NULL) {
	  traceEvent(TRACE_WARNING, "Wrong engine specified (-E flag): see help.");
	} else
	  readOnlyGlobals.engineId = (u_int8_t)atoi(theItem);
      }
      break;
    case 'e':
      readOnlyGlobals.flowExportDelay = atoi(optarg);
      if(readOnlyGlobals.flowExportDelay >= 1000) {
	/*
	  NOTE

	  A value of 1000 or more ms will not allow nanosleep to
	  operate properly as 1000 ms = 1 sec and nanosleep does
	  not accept ms sleeps of 1000 or more ms
	*/
	traceEvent(TRACE_WARNING, "Maximum flow export delay is 999 ms");
	readOnlyGlobals.flowExportDelay = 999;
      }
      break;
    case 'g':
      readOnlyGlobals.pidPath = strdup(optarg);
      break;
    case 'O':
      readOnlyGlobals.numProcessThreads = atoi(optarg);
      if(readOnlyGlobals.numProcessThreads > MAX_NUM_PCAP_THREADS) {
	traceEvent(TRACE_ERROR, "You can spawn at most %d threads.",
		   MAX_NUM_PCAP_THREADS);
	readOnlyGlobals.numProcessThreads = MAX_NUM_PCAP_THREADS;
      }

      if(readOnlyGlobals.numProcessThreads <= 0) readOnlyGlobals.numProcessThreads = 1;
      break;
    case 'f':
      if((optarg[0] == '\"') || (optarg[0] == '\'')) {
	readOnlyGlobals.netFilter = strdup(&optarg[1]);
	readOnlyGlobals.netFilter[strlen(readOnlyGlobals.netFilter)-2] = '\0';
      } else {
	readOnlyGlobals.netFilter = strdup(optarg);
      }
      break;
    case 'F':
      readOnlyGlobals.file_dump_timeout = atoi(optarg);
      if(readOnlyGlobals.file_dump_timeout < 60) {
	readOnlyGlobals.file_dump_timeout = 60;
	traceEvent(TRACE_WARNING, "Sorry: -F cannot be set below 60 (sec)");
      }
      break;
    case 'h':
      usage();
      return(-1);
    case 'i':
      {
#ifdef WIN32
	struct stat statbuf;

	if(!strcmp(optarg, "none"))
	  readOnlyGlobals.captureDev = strdup(optarg);
	else if(stat(optarg, &statbuf) != 0) {
	  readOnlyGlobals.captureDev = printAvailableInterfaces(optarg);
	} else
#endif
	  {
	    if(readOnlyGlobals.captureDev != NULL) free(readOnlyGlobals.captureDev);
	    readOnlyGlobals.captureDev = strdup(optarg);
	  }
      }
      break;
    case 'm':
      readOnlyGlobals.minNumFlowsPerPacket = atoi(optarg);
      break;
    case 'p':
      {
	int a, b, c, d, e, f;

	if(sscanf(optarg, "%d/%d/%d/%d/%d/%d", &a, &b, &c, &d, &e, &f) != 6) {
	  traceEvent(TRACE_WARNING, "Sorry: the -p parameter has an invalid format");
	} else {
	  readOnlyGlobals.ignoreVlan     = (a == 0) ? 1 : 0;
	  readOnlyGlobals.ignoreProtocol = (b == 0) ? 1 : 0;
	  readOnlyGlobals.ignoreIP       = (c == 0) ? 1 : 0;
	  readOnlyGlobals.ignorePorts    = (d == 0) ? 1 : 0;
	  readOnlyGlobals.ignoreTos      = (e == 0) ? 1 : 0;

	  if(f == 0) {
#ifdef HAVE_GEOIP
	    if(readOnlyGlobals.geo_ip_asn_db != NULL)
	      GeoIP_delete(readOnlyGlobals.geo_ip_asn_db);
	    readOnlyGlobals.geo_ip_asn_db = NULL;
#endif
	  }
	}
      }
      break;
    case 'r':
      if(readOnlyGlobals.numLocalNetworks == 0) {
	traceEvent(TRACE_WARNING, "Ignored -r: it must be specified after -L");
      } else
	readOnlyGlobals.setLocalTrafficDirection = 1;
      break;
    case 's':
      readOnlyGlobals.scanCycle = atoi(optarg);
      if(readOnlyGlobals.scanCycle <= 0) readOnlyGlobals.scanCycle = 1;
      break;
#ifndef WIN32
    case 'G':
      readOnlyGlobals.becomeDaemon = 1;
      break;
#endif
    case 'l':
      readOnlyGlobals.sendTimeout = atoi(optarg);
      break;
    case 'L':
      parseLocalAddressLists(optarg);
      break;
    case 'M':
      readOnlyGlobals.maxNumActiveFlows = (u_int)atoi(optarg);
      break;
    case 'S':
      {
	u_int a, b, begin = 0;

	if(optarg[0] == '@')
	  readOnlyGlobals.fakePktSampling = 1, begin = 1;

	if(sscanf(&optarg[begin], "%u:%u", &a, &b) == 2) {
	  readOnlyGlobals.pktSampleRate = a;
	  readOnlyGlobals.flowSampleRate = b;
	} else {
	  traceEvent(TRACE_WARNING, "Unable to parse sampling option: discarded");
	  readOnlyGlobals.pktSampleRate = 1;
	  readOnlyGlobals.flowSampleRate = 1;
	}

	if(readOnlyGlobals.pktSampleRate > MAX_SAMPLE_RATE) {
	  readOnlyGlobals.pktSampleRate = MAX_SAMPLE_RATE;
	  traceEvent(TRACE_WARNING, "Packet sample rate set to %d [range 1:%d]",
		     MAX_SAMPLE_RATE, MAX_SAMPLE_RATE);
	}
	if(readOnlyGlobals.pktSampleRate == 0) readOnlyGlobals.pktSampleRate = 1;

	if(readOnlyGlobals.flowSampleRate > MAX_SAMPLE_RATE) {
	  readOnlyGlobals.flowSampleRate = MAX_SAMPLE_RATE;
	  traceEvent(TRACE_WARNING, "Flow sample rate set to %d [range 1:%d]",
		     MAX_SAMPLE_RATE, MAX_SAMPLE_RATE);
	}
	if(readOnlyGlobals.flowSampleRate == 0) readOnlyGlobals.flowSampleRate = 1;
      }
      break;
    case 't':
      readOnlyGlobals.lifetimeTimeout = atoi(optarg);
      if(readOnlyGlobals.lifetimeTimeout == 0) {
	readOnlyGlobals.lifetimeTimeout = 1;
	traceEvent(TRACE_WARNING, "Minimum flow lifetime can't be set to zero: set to %d sec",
		   readOnlyGlobals.lifetimeTimeout);
      }
      break;
    case 'u':
      id = atoi(optarg);
      if(id > (u_int16_t)-1)
	traceEvent(TRACE_WARNING, "The -%c value %s it out of boundaries (0...%u) and it will be truncated",
		   opt, optarg, (u_int16_t)-1);
      readOnlyGlobals.inputInterfaceIndex = id;
      break;
    case 'z':
      readOnlyGlobals.minFlowSize = (u_int)atoi(optarg);
      break;
#ifdef HAVE_PF_RING
    case 'Z':
      readOnlyGlobals.cluster_id = atoi(optarg);
      break;
#endif
    case 'v':
      probeVersion();
      exit(0);
    case 'w':
      readOnlyGlobals.flowHashSize = atoi(optarg);
      if(readOnlyGlobals.flowHashSize < MIN_HASH_SIZE) {
	readOnlyGlobals.flowHashSize = MIN_HASH_SIZE;
	traceEvent(TRACE_INFO, "Minimum hash size if %d.",
		   readOnlyGlobals.flowHashSize);
      }
      readOnlyGlobals.hostHashSize = readOnlyGlobals.flowHashSize/2;
      break;

    case 'W':
      readOnlyGlobals.disableIPv6 = 1;
      break;

#ifndef WIN32
    case 'I':
      {
	u_int len = strlen(optarg), max_len = sizeof(readOnlyGlobals.nprobeId)-1;

	if(len >= max_len) len = max_len;
	strncpy(readOnlyGlobals.nprobeId, optarg, len);
	readOnlyGlobals.nprobeId[len] = '\0';
	readOnlyGlobals.useSyslog = 1;
      }
      break;
#endif
    case 'n':
      if(strcmp(optarg, "none")) {
	if(readOnlyGlobals.none_specified) {
	  traceEvent(TRACE_WARNING, "-n <host:port> is ignored as '-n none' has beed specified before");
	} else {
	  char *port = NULL, *addr = NULL;

	  opt_n = 1;

	  port = strrchr(optarg, ':');

	  if(port != NULL) {
	    port[0] = '\0';
	    port++;
	    addr =  optarg;

	    if(addr[0] == '[') {
	      /*
		IPv6 addresses should be delimited by square brackets
		according to RFC 2732.
	      */
	      addr++;

	      if(strlen(addr) > 0)
		addr[strlen(addr)-1] = '\0';
	    }

	    if(initNetFlow(addr, atoi(port)) == 0)
	      mandatoryParamOk++;
	  } else {
	    usage();
	  }
	}
      } else {
	if(readOnlyGlobals.numCollectors > 0) {
	  traceEvent(TRACE_WARNING, "'-n none' is ignored as '-n <host:port>' has beed specified before");
	} else {
	  readOnlyGlobals.none_specified = 1, mandatoryParamOk++;
	}
      }
      break;
    case 'N':
      traceEvent(TRACE_WARNING, "This option is obsolete and will be removed");
      break;
    case 'o':
      readOnlyGlobals.templatePacketsDelta = (u_short)atoi(optarg);
      break;
#ifdef IP_HDRINCL
    case 'q':
      {
	if(opt_n == 1) {
	  traceEvent(TRACE_ERROR,
		     "You need to specify the --sender-address|-q option before the --collector|-n option."
		     " Please try again.");
	  exit(0);
	}

	readOnlyGlobals.bindAddr = strtok(optarg, ":");
	if(readOnlyGlobals.bindAddr != NULL) {
	  readOnlyGlobals.bindAddr = strdup(readOnlyGlobals.bindAddr);
	  readOnlyGlobals.bindPort = strtok(NULL, ":");
	  if(readOnlyGlobals.bindPort == NULL)
	    usage();
	  else
	    readOnlyGlobals.bindPort = strdup(readOnlyGlobals.bindPort);
	} else
	  usage();

	if(readOnlyGlobals.bindAddr != NULL) {
	  memset(&readOnlyGlobals.sockIn, 0, sizeof(readOnlyGlobals.sockIn));
	  /*
	    FreeBSD only
	    readOnlyGlobals.sockIn.sin_len = sizeof(struct sockaddr_in);
	  */
	  readOnlyGlobals.sockIn.sin_family = AF_INET6;

	  if(readOnlyGlobals.bindPort)
	    readOnlyGlobals.sockIn.sin_port   = (int)htons((unsigned short int)atoi(readOnlyGlobals.bindPort));

	  if(!inet_aton(readOnlyGlobals.bindAddr, &readOnlyGlobals.sockIn.sin_addr)) {
	    traceEvent(TRACE_ERROR, "Unable to convert address '%s'. "
		       "Not binding to a particular interface", readOnlyGlobals.bindAddr);
	    readOnlyGlobals.sockIn.sin_addr.s_addr = INADDR_ANY;
	  }

	  /*
	    If we ask to bind to IPv4 via -q then we
	    implicitly ask to use IPv4
	  */
	  if(strstr(readOnlyGlobals.bindAddr, ":") == NULL)
	    readOnlyGlobals.useIpV6 = 0;
	}
      }
      break;
#endif

    case 'Q':
      id = atoi(optarg);
      if(id > (u_int16_t)-1)
	traceEvent(TRACE_WARNING, "The -%c value %s it out of boundaries and it will be truncated",
		   opt, optarg);
      readOnlyGlobals.outputInterfaceIndex = (u_int16_t)id;
      break;

    case 'T':
      {
	u_int8_t ignore_template = 1;

	if(reparse_options) {
	  if(strcmp(readOnlyGlobals.baseTemplateBufferV4, optarg) != 0) {
	    /* Template is NOT the same */
	    ignore_template = 0;
	  }
	}

	if((readOnlyGlobals.baseTemplateBufferV4 == NULL) || (!ignore_template)) {
	  readOnlyGlobals.baseTemplateBufferV4 = strdup(optarg);
	  if(readOnlyGlobals.netFlowVersion == 5) readOnlyGlobals.netFlowVersion = 9; /* NetFlow v9 */
	  if(readOnlyGlobals.useNetFlow == 0xFF) readOnlyGlobals.useNetFlow = 1;

	  if(reparse_options)
	    compileTemplates(1);
	}
      }
      break;

    case 'U':
      {
	u_int v = atoi(optarg);

	if(v < 256) {
	  traceEvent(TRACE_WARNING, "Please use templatedId >= 256");
	} else {
	  readOnlyGlobals.idTemplate = v;
	  if(readOnlyGlobals.netFlowVersion != 9) readOnlyGlobals.netFlowVersion = 9; /* NetFlow v9 */
	  if(readOnlyGlobals.useNetFlow == 0xFF) readOnlyGlobals.useNetFlow = 1;
	}
      }
      break;

    case 'V':
      readOnlyGlobals.netFlowVersion = atoi(optarg);
      if((readOnlyGlobals.netFlowVersion != 5)
	 && (readOnlyGlobals.netFlowVersion != 9)
	 && (readOnlyGlobals.netFlowVersion != 10)) {
	traceEvent(TRACE_ERROR,
		   "nProbe supports 5 (NetFlow 5), 9 (NetFlow 9) and 10 (IPFIX)");
	exit(0);
      }
      break;

    case 249: /* --log <low port>:<num ports> */
      if(sscanf(optarg, "%u:%u", &readOnlyGlobals.logPluginLowPort,
		&readOnlyGlobals.logPluginNumPorts) != 2) {
	readOnlyGlobals.logPluginLowPort = atoi(optarg), readOnlyGlobals.logPluginNumPorts = 1;
      }

      if(readOnlyGlobals.logPluginNumPorts > 32)
	readOnlyGlobals.logPluginNumPorts = 32;
      else if(readOnlyGlobals.logPluginNumPorts == 0)
	readOnlyGlobals.logPluginNumPorts = 1;

      readOnlyGlobals.logPluginEnabled = 1;
      break;

    case 250: /* --nfLitePlugin <low port>:<num ports> */
      if(sscanf(optarg, "%u:%u", &readOnlyGlobals.nfLitePluginLowPort,
		&readOnlyGlobals.nfLitePluginNumPorts) != 2) {
	readOnlyGlobals.nfLitePluginLowPort = atoi(optarg), readOnlyGlobals.nfLitePluginNumPorts = 1;
      }

      if(readOnlyGlobals.nfLitePluginNumPorts > 32)
	readOnlyGlobals.nfLitePluginNumPorts = 32;
      else if(readOnlyGlobals.nfLitePluginNumPorts == 0)
	readOnlyGlobals.nfLitePluginNumPorts = 1;

#ifdef IP_HDRINCL
      traceEvent(TRACE_NORMAL, "IMPORTANT: Enabling NflitePlugin will also enable IP address forging, thus");
      traceEvent(TRACE_NORMAL, "IMPORTANT: flows appear as they were sent from the NflitePlugin-enabled switch");
#endif
      readOnlyGlobals.nfLitePluginEnabled = 1;
      break;

    case 244:
      free(readOnlyGlobals.unprivilegedUser);
      readOnlyGlobals.unprivilegedUser = strdup(optarg);
      break;

    case 245:
      readOnlyGlobals.disableFlowCache = 1;
      break;

    case 246:
      readOnlyGlobals.fakePacketCapture = 1;
      break;

    case 247:
      readOnlyGlobals.quick_mode = 1;
      break;

    case 248:
      readOnlyGlobals.tracePerformance = 1;
      break;

    case 254:
      readOnlyGlobals.enable_debug = 1;
      break;

    default:
      //usage();
      break;
    }
  }

  if(reparse_options) return(0);

  if(unlikely(readOnlyGlobals.enable_debug)) {
    // readOnlyGlobals.numProcessThreads = 1;
  }

  if(readOnlyGlobals.outputInterfaceIndex == 0)
    traceEvent(TRACE_WARNING, "The output interfaceId is set to 0: did you forget to use -Q perhaps ?");

  if(readOnlyGlobals.inputInterfaceIndex == 0)
    traceEvent(TRACE_WARNING, "The input interfaceId is set to 0: did you forget to use -u perhaps ?");

  /*
    This guarantees that the hash size is a multiple of the number of threads
    hence that we do not need locks in the hash when using lockless hash
  */
  readOnlyGlobals.flowHashSize -= (readOnlyGlobals.flowHashSize % readOnlyGlobals.numProcessThreads);

  readOnlyGlobals.maxExportQueueLen = max(readOnlyGlobals.maxExportQueueLen, readOnlyGlobals.flowHashSize);

  if((readOnlyGlobals.captureDev != NULL)
     && (readOnlyGlobals.pcapFileList != NULL)) {
    traceEvent(TRACE_NORMAL, "-i is ignored as --pcap-file-list has been used");
    free(readOnlyGlobals.captureDev);
    readOnlyGlobals.captureDev = NULL;
  }

  if(readOnlyGlobals.userTemplateBuffer.v9TemplateElementList[0] != NULL) {
    readOnlyGlobals.scanCycle = readOnlyGlobals.file_dump_timeout;
    traceEvent(TRACE_INFO, "Scan cycle (-s) set to the -F value: %d",
	       readOnlyGlobals.scanCycle);
  }

  if(readOnlyGlobals.useNetFlow == 0xFF) readOnlyGlobals.useNetFlow = 1;

  if(readOnlyGlobals.netFlowVersion == 5) {
    if(readOnlyGlobals.minNumFlowsPerPacket == (u_short)-1)
      readOnlyGlobals.minNumFlowsPerPacket = V5FLOWS_PER_PAK; /* Default */

    if(readOnlyGlobals.minNumFlowsPerPacket > V5FLOWS_PER_PAK) {
      traceEvent(TRACE_WARNING,
		 "Sorry: the min # of flows per packet (%d) cannot be set over %d",
		 readOnlyGlobals.minNumFlowsPerPacket, V5FLOWS_PER_PAK);
      readOnlyGlobals.minNumFlowsPerPacket = V5FLOWS_PER_PAK;
    }

    readOnlyGlobals.disableIPv6 = 1;
  }

  traceEvent(TRACE_NORMAL, "Welcome to nprobe v.%s (%s) for %s %s",
	     version, nprobe_revision, osName,
#ifdef HAVE_PF_RING
	     "with native PF_RING acceleration"
#else
	     ""
#endif
	     );

  if(((fd = fopen("nprobe.license", "r")) != NULL)
     || ((fd = fopen("/etc/nprobe.license", "r")) != NULL)) {
    char license[256] = { 0 }, *ret;

    ret = fgets(license, sizeof(license), fd);
    fclose(fd);

    traceEvent(TRACE_NORMAL, "nProbe License: %s", license);
  }

  if(readOnlyGlobals.traceMode) traceEvent(TRACE_INFO, "Tracing enabled");

  if(readWriteGlobals->num_src_mac_export > 0) {
    int i;

    for(i = 0; i<readWriteGlobals->num_src_mac_export; i++)
      traceEvent(TRACE_INFO,
		 "Using interface idx %d for flows originating from %02X:%02X:%02X:%02X:%02X:%02X",
		 readOnlyGlobals.mac_if_match[i].interface_id,
		 readOnlyGlobals.mac_if_match[i].mac_address[0],
		 readOnlyGlobals.mac_if_match[i].mac_address[1],
		 readOnlyGlobals.mac_if_match[i].mac_address[2],
		 readOnlyGlobals.mac_if_match[i].mac_address[3],
		 readOnlyGlobals.mac_if_match[i].mac_address[4],
		 readOnlyGlobals.mac_if_match[i].mac_address[5]);
  }

  if(readOnlyGlobals.dirPath) {
    struct stat statbuf;

    if((stat(readOnlyGlobals.dirPath, &statbuf) != 0)
       || (!(statbuf.st_mode & S_IFDIR)) /* It's not a directory */
       || (!(statbuf.st_mode & S_IWRITE)) /* It's not writable    */
       ) {
      traceEvent(TRACE_ERROR,
		 "Sorry, the path you specified with -P is invalid.");
      traceEvent(TRACE_ERROR,
		 "Make sure the directory exists and it's writable.");
      exit(-1);
    }

    readWriteGlobals->flowFd = NULL;
    mandatoryParamOk = 1; /* -P can substitute -n */
    traceEvent(TRACE_NORMAL, "Dumping flow files every %d sec into directory %s",
	       readOnlyGlobals.file_dump_timeout, readOnlyGlobals.dirPath);
  }

  if((readOnlyGlobals.numCollectors == 0) && (!readOnlyGlobals.none_specified)) {
    traceEvent(TRACE_WARNING, "-n parameter is missing. 127.0.0.1:2055 will be used.\n");
    initNetFlow("127.0.0.1", 2055);
    mandatoryParamOk = 1;
  }

#if 0
  if(readOnlyGlobals.disableFlowCache
     && ((readOnlyGlobals.numCollectors == 0) || readOnlyGlobals.none_specified)) {
    traceEvent(TRACE_WARNING, "Enabling flow cache as the probe is not used in collector/proxy mode");
    readOnlyGlobals.disableFlowCache = 0;
  }
#endif

  if(!mandatoryParamOk) {
    usage();
    return(-1);
  }

  if(readOnlyGlobals.netFlowVersion == 10)
    fixTemplatesToIPFIX();

#ifdef linux
  setCpuAffinity(readOnlyGlobals.captureDev, readOnlyGlobals.cpuAffinity);
#endif

  return(0);
}

/* ****************************************************** */

static char *printPayloadValue(u_char payloadExportType) {
  switch(payloadExportType) {
  case 0:
    return("no payload");
    break;
  case 1:
    return("full payload");
    break;
  case 2:
    return("payload only with SYN set");
    break;
  default:
    return("??");
  }
}

/* ****************************************************** */

static void stopCaptureFlushAll(void) {
  u_int hash_idx = 0;

  readWriteGlobals->stopPacketCapture = 1;
  traceEvent(TRACE_INFO, "nProbe is shutting down...");

#ifdef HAVE_PF_RING
  if(1)
  if(readWriteGlobals->ring) {
    int num = 0;
    pfring_breakloop(readWriteGlobals->ring);
    traceEvent(TRACE_NORMAL, "Waiting for PF_RING termination");

    while(readWriteGlobals->ring_enabled) {
      if(++num == 3) 
	break;
      else
	sleep(1);
    }

    traceEvent(TRACE_NORMAL, "PF_RING terminated");

    pfring_close(readWriteGlobals->ring);
    readWriteGlobals->ring = NULL;
  }
#endif

  readWriteGlobals->shutdownInProgress = 1;

  /* Expedite export */
  readOnlyGlobals.flowExportDelay = 0;

  traceEvent(TRACE_INFO, "Exporting pending buckets...\n");
  for(hash_idx=0; hash_idx<readOnlyGlobals.numProcessThreads; hash_idx++) {
    walkHash(hash_idx, 1);
  }

  if(readWriteGlobals->exportBucketsLen > 0) {
    traceEvent(TRACE_INFO, "Waiting to export queued buckets... [queue len=%d]",
	       readWriteGlobals->exportBucketsLen);

    while(readWriteGlobals->exportBucketsLen > 0) {
      signalCondvar(&readWriteGlobals->exportQueueCondvar, 0);
      ntop_sleep(1);
      traceEvent(TRACE_NORMAL, "Still %d queued buckets to be exported...",
		 readWriteGlobals->exportBucketsLen);
    }
  }

  checkNetFlowExport(1);
  traceEvent(TRACE_INFO, "Pending buckets have been exported...\n");
}

/* ****************************************************** */

void term_pcap(pcap_t **p) {
  struct pcap_stat pcapStat;

  if(p == NULL) return;

  if(readOnlyGlobals.numProcessThreads > 1)
    pthread_rwlock_wrlock(&readWriteGlobals->pcapLock);

  if(pcap_stats(*p, &pcapStat) >= 0)
    traceEvent(TRACE_NORMAL, "Packet stats: %u/%u pkts rcvd/dropped",
	       pcapStat.ps_recv, pcapStat.ps_drop);

  if(readOnlyGlobals.pcapFile) {
    pcap_close(*p);
    /*
      No clue why sometimes it crashes
      so we free only when reading .pcap dump files
    */
    free(readOnlyGlobals.pcapFile);
  }

  *p = NULL;

  /* No unlock */
}

/* ****************************************************** */

void shutdown_nprobe(void) {
  static u_char once = 0;
  FlowHashBucket *list;
  u_int i, bucketsAllocated;

  if(once) return; else once = 1;

  stopCaptureFlushAll();

  // ntop_sleep(1);
  signalCondvar(&readWriteGlobals->exportQueueCondvar, 0);

  if(readOnlyGlobals.dequeueBucketToExport_up)
    waitCondvar(&readWriteGlobals->termCondvar); /* Wait until dequeueBucketToExport() ends */

  traceEvent(TRACE_INFO, "Flushing queued flows...\n");
  checkNetFlowExport(1 /* force export */);

  traceEvent(TRACE_INFO, "Freeing memory...\n");

  for(i = 0; i<readOnlyGlobals.numCollectors; i++)
    close(readOnlyGlobals.netFlowDest[i].sockFd);

  close_dump_file();

  free_bitmask(&readOnlyGlobals.udpProto);
  free_bitmask(&readOnlyGlobals.tcpProto);

  unload_mappings();

  if(readOnlyGlobals.pcapPtr) {
    printPcapStats(readOnlyGlobals.pcapPtr);
    term_pcap(&readOnlyGlobals.pcapPtr);
    readOnlyGlobals.pcapPtr = NULL;
  }

  for(i=0; i<readOnlyGlobals.numProcessThreads; i++)
    free(readWriteGlobals->theFlowHash[i]);

  freeHostHash();

  if(readOnlyGlobals.captureDev != NULL) free(readOnlyGlobals.captureDev);

  for(i=0; i<readOnlyGlobals.numActiveTemplates; i++) {
    if(readOnlyGlobals.templateBuffers[i].buffer)
      free(readOnlyGlobals.templateBuffers[i].buffer);
  }

  list = readWriteGlobals->exportQueue;

  while(list != NULL) {
    FlowHashBucket *nextEntry = list->core.hash.next;

    if(list->ext->extensions != NULL) {
      if(list->ext->extensions->mplsInfo != NULL) free(list->ext->extensions->mplsInfo);
      free(list->ext);
    }

    free(list);
    list = nextEntry;
  }

  for(i=0; i<NUM_FRAGMENT_LISTS; i++) {
    IpV4Fragment *list = readWriteGlobals->fragmentsList[i];

    while(list != NULL) {
      IpV4Fragment *next = list->next;
      free(list);
      list = next;
    }
  }

#ifdef DEBUG
  if(readWriteGlobals->bucketsAllocated > 0)
    traceEvent(TRACE_INFO, "WARNING ===> readWriteGlobals->bucketsAllocated: %d\n",
	       readWriteGlobals->bucketsAllocated);
#endif

#ifdef HAVE_FASTBIT
  term_fastbit(readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].v9TemplateElementList);
#endif

#ifndef WIN32
  if(readOnlyGlobals.useSyslog)
    closelog();
#endif

  termPlugins();

  if(readOnlyGlobals.argv) {
    for(i=0; i<readOnlyGlobals.argc; i++)
      free(readOnlyGlobals.argv[i]);

    free(readOnlyGlobals.argv);
  }

  /* Clean threads */
#if 0
  traceEvent(TRACE_INFO, "Cleaning threads");
  pthread_exit(&readWriteGlobals->walkHashThread);
  pthread_exit(&readWriteGlobals->dequeueThread);
#endif

  for(bucketsAllocated=0, i=0; i<readOnlyGlobals.numProcessThreads; i++)
    bucketsAllocated += readWriteGlobals->bucketsAllocated[i];
  traceEvent(TRACE_INFO, "Still allocated %u hash buckets", bucketsAllocated);

  printProcessingStats();

  /* Clean globals */
  traceEvent(TRACE_INFO, "Cleaning globals");

  free(readOnlyGlobals.csv_separator);
  free(readOnlyGlobals.dirPath);
  free(readOnlyGlobals.unprivilegedUser);

  // free(readOnlyGlobals.packetProcessThread);
  if(readOnlyGlobals.baseTemplateBufferV4) free(readOnlyGlobals.baseTemplateBufferV4);
  if(readOnlyGlobals.stringTemplateV4) free(readOnlyGlobals.stringTemplateV4);
  if(readOnlyGlobals.stringTemplateV6) free(readOnlyGlobals.stringTemplateV6);

  if(unlikely(readOnlyGlobals.tracePerformance)) {
    printProcessingStats();
  }

  traceEvent(TRACE_INFO, "nProbe terminated.");
  dumpLogEvent(probe_stopped, severity_info, "nProbe stopped");
  if(readOnlyGlobals.eventLogPath) free(readOnlyGlobals.eventLogPath);

  free(readWriteGlobals); /* Do not move it up as it's needed for logging */

#ifndef WIN32
  endpwent();
#endif

  exit(0);
}

/* ******************************************* */

static int openDevice(char ebuf[], int printErrors, char *pcapFilePath) {
  u_char open_device = 1;

  if(readOnlyGlobals.enableHttpPlugin || readOnlyGlobals.enableDnsPlugin
     || readOnlyGlobals.enableMySQLPlugin || readOnlyGlobals.enableSipPlugin
     || readOnlyGlobals.enableOraclePlugin
     || readOnlyGlobals.enableGtpPlugin
     || readOnlyGlobals.enableRadiusPlugin
     || readOnlyGlobals.enableL7BridgePlugin
     || readOnlyGlobals.enable_l7_protocol_discovery)
    readOnlyGlobals.snaplen = PCAP_LONG_SNAPLEN;

  traceEvent(TRACE_NORMAL, "Using packet capture length %u", readOnlyGlobals.snaplen);

  if((readOnlyGlobals.captureDev != NULL)
     && (strcmp(readOnlyGlobals.captureDev, "none") == 0)) {
    readOnlyGlobals.do_not_drop_privileges = 1;
    return(0);
  }

  if(readOnlyGlobals.captureDev != NULL) {
    /* Try if the passed device is instead a dump file */

#ifdef HAVE_PF_RING
    if(strstr(readOnlyGlobals.captureDev, ".metadata")) {
      if((readOnlyGlobals.metadata_fd = fopen(readOnlyGlobals.captureDev, "rb")) == NULL) {
	traceEvent(TRACE_ERROR, "Unable to open file %s", readOnlyGlobals.captureDev);
	return(-1);
      } else {
	traceEvent(TRACE_NORMAL, "Open metadata file %s", readOnlyGlobals.captureDev);
	goto do_datalink;
      }
    }
#endif

    readOnlyGlobals.pcapPtr = pcap_open_offline(readOnlyGlobals.captureDev, ebuf);
    if(readOnlyGlobals.pcapPtr != NULL) {
      readOnlyGlobals.pcapFile = strdup(readOnlyGlobals.captureDev);
      readOnlyGlobals.snaplen = 65535;
    }
  } else if(pcapFilePath != NULL) {
    if(readOnlyGlobals.pcapPtr != NULL) {
      term_pcap(&readOnlyGlobals.pcapPtr);
      readOnlyGlobals.pcapPtr = NULL;
    }

    readOnlyGlobals.pcapPtr = pcap_open_offline(pcapFilePath, ebuf);
    if(readOnlyGlobals.pcapPtr != NULL) {
      traceEvent(TRACE_NORMAL, "Processing packets from file %s", pcapFilePath);
      readOnlyGlobals.pcapFile = strdup(pcapFilePath);
      readOnlyGlobals.snaplen = LONG_SNAPLEN;
    } else
      return(-1);
  } else
    readOnlyGlobals.pcapPtr = NULL;

  if(readOnlyGlobals.pcapPtr == NULL) {
    /* Find the default device if not specified */
    if(readOnlyGlobals.captureDev == NULL) {
#ifdef WIN32
      readOnlyGlobals.captureDev = printAvailableInterfaces(NULL);
#else
      readOnlyGlobals.captureDev = pcap_lookupdev(ebuf);
#endif
      if(readOnlyGlobals.captureDev == NULL) {
	if(printErrors)
	  traceEvent(TRACE_ERROR,
		     "Unable to locate default interface (%s)\n", ebuf);
	return(-1);
      } else {
	char *_captureDev = strdup(readOnlyGlobals.captureDev);
	readOnlyGlobals.captureDev = _captureDev;
      }
    }

#ifdef HAVE_PF_RING
    readWriteGlobals->ring = open_ring(readOnlyGlobals.captureDev, &open_device, 0);
#endif

    if(open_device) {
      readOnlyGlobals.pcapPtr = pcap_open_live(readOnlyGlobals.captureDev,
					       readOnlyGlobals.snaplen,
					       readOnlyGlobals.promisc_mode /* promiscuous mode */,
					       1000 /* ms */,
					       ebuf);

      if(readOnlyGlobals.pcapPtr == NULL)  {
	if(printErrors)
	  traceEvent(TRACE_ERROR, "Unable to open interface %s.\n", readOnlyGlobals.captureDev);

#ifndef WIN32
	if((getuid () && geteuid ()) || setuid (0)) {
	  if(printErrors) {
	    traceEvent(TRACE_ERROR, "nProbe opens the network interface "
		       "in promiscuous mode, ");
	    traceEvent(TRACE_ERROR, "so it needs root permission "
		       "to run. Quitting...");
	  }
	}
#endif
	return(-1);
      }
    }
  }

#ifdef HAVE_PF_RING
  if(readWriteGlobals->ring != NULL) {
  do_datalink:
    readOnlyGlobals.datalink = DLT_EN10MB;
  } else
#endif
    readOnlyGlobals.datalink = pcap_datalink(readOnlyGlobals.pcapPtr);

  /* ************************ */

  if(readOnlyGlobals.netFilter != NULL) {
#ifdef HAVE_PF_RING
    if(readWriteGlobals->ring != NULL) {
      ; /* Nothing to do the PF_RING modules does the magic */
    } else
#endif
    if(readOnlyGlobals.pcapPtr == NULL) {
      traceEvent(TRACE_ERROR, "Unable to set filter %s: perhaps are you running in native PF_RING mode?",
		 readOnlyGlobals.netFilter);
    } else {
      struct bpf_program fcode;
      struct in_addr netmask;

      netmask.s_addr = htonl(0xFFFFFF00);

      if((pcap_compile(readOnlyGlobals.pcapPtr, &fcode, readOnlyGlobals.netFilter, 1, netmask.s_addr) < 0)
	 || (pcap_setfilter(readOnlyGlobals.pcapPtr, &fcode) < 0)) {
	if(printErrors)
	  traceEvent(TRACE_ERROR, "Unable to set filter %s. Filter ignored.\n", readOnlyGlobals.netFilter);
	/* return(-1); */
      } else {
	if(printErrors)
	  traceEvent(TRACE_INFO, "Packet capture filter set to \"%s\"",
		     readOnlyGlobals.netFilter);
      }
    }

    free(readOnlyGlobals.netFilter);
  }

  return(0);
}

/* ****************************************************** */

static int restoreInterface(char ebuf[]) {
  if(readOnlyGlobals.pcapFile == NULL) {
    int rc = -1;

    traceEvent(TRACE_INFO,
	       "Error while capturing packets: %s",
	       pcap_geterr(readOnlyGlobals.pcapPtr));
    traceEvent(TRACE_INFO, "Waiting until the interface comes back...");

    while(rc == -1) {
      ntop_sleep(1);
      rc = openDevice(ebuf, 0, NULL);
    }

    traceEvent(TRACE_INFO, "The interface is now awailable again.");
    return(rc);
  }

  return(-2);
}

/* ****************************************************** */

#ifndef HAVE_PCAP_NEXT_EX
int pcap_next_ex(pcap_t *p, struct pcap_pkthdr **pkt_header,
                 const u_char **pkt_data)
{
  static struct pcap_pkthdr h;

  (*pkt_data) = pcap_next(p, &h);
  (*pkt_header) = &h;
  if(*pkt_data)
    return(1);
  else
    return(0);
}
#endif

/* ****************************************************** */

u_int32_t usecdiff(const struct timeval *last, const struct timeval *prev) {
  struct timeval res;

  res.tv_sec = last->tv_sec - prev->tv_sec;
  if(prev->tv_usec > last->tv_usec) {
    res.tv_usec = last->tv_usec + 1000000 - prev->tv_usec;
    res.tv_sec--;
  } else
    res.tv_usec = last->tv_usec - prev->tv_usec;

  return(res.tv_sec*1000000 + res.tv_usec);
}

/* ****************************************************** */

inline int next_pcap_packet(pcap_t *p, struct pcap_pkthdr *h, const u_char *pkt_data) {
  int rc;
  u_char *pkt;
  struct pcap_pkthdr *hdr;

  if(readOnlyGlobals.numProcessThreads > 1)
    pthread_rwlock_wrlock(&readWriteGlobals->pcapLock);

  rc = pcap_next_ex(readOnlyGlobals.pcapPtr, &hdr, (const u_char**)&pkt);
  if((rc > 0) && (pkt != NULL) && (hdr->caplen > 0)) {
    hdr->caplen = min(hdr->caplen, readOnlyGlobals.snaplen);
    memcpy(h, hdr, sizeof(struct pcap_pkthdr)),
      memcpy((void*)pkt_data, (const void*)pkt, h->caplen);
  }

  if(unlikely(readOnlyGlobals.enable_debug)) {
    if(readWriteGlobals->currentPkts[0] > 0) {
      static struct timeval last;

      if(last.tv_sec != 0) {
	u_int32_t us = usecdiff(&h->ts, &last);

	/* traceEvent(TRACE_INFO, "Sleeping %u usec", us); */
        // usleep(us); FIX
      }

      memcpy(&last, &h->ts, sizeof(last));
    }

    gettimeofday(&h->ts, NULL); /* Use current time */
  }

  if(readOnlyGlobals.numProcessThreads > 1)
    pthread_rwlock_unlock(&readWriteGlobals->pcapLock);

  return(rc);
}

/* ****************************************************** */

static void* processPackets(void* _thid) {
  unsigned long thread_id = (unsigned long)_thid;
  PacketQueue *queue = &readWriteGlobals->packetQueues[thread_id];
  QueuedPacket *slot;

  if(readOnlyGlobals.numProcessThreads == 1) {
    /* Sanity check */
    traceEvent(TRACE_WARNING, "Internal error: processPackets() called in single thread");
    return(NULL);
  }

  // setThreadAffinity(thread_id);

  while(!readWriteGlobals->shutdownInProgress) {
    if(queue->num_insert != queue->num_remove) {
      slot = &queue->queueSlots[queue->remove_idx];

      if(slot->packet_ready) {
	deepPacketDecode(thread_id,
			 slot->packet_if_idx,
			 &slot->h, slot->p,
			 0 /* sampledPacket */,
			 1 /* numPkts */,
			 NO_INTERFACE_INDEX, NO_INTERFACE_INDEX,
			 0 /* flow_sender_ip */,
			 slot->packet_hash);

	slot->packet_ready = 0, queue->remove_idx = (queue->remove_idx + 1) % DEFAULT_QUEUE_CAPACITY, queue->num_remove++;
	continue;
      }
    }

    // waitCondvar(&queue->dequeue_condvar);
    usleep(1);
  }

  return(NULL);
}

/* ****************************************************** */

static void* fetchPackets(void* _thid) {
  char ebuf[PCAP_ERRBUF_SIZE];
  const u_char *packet;
  u_short packetToGo = readOnlyGlobals.fakePktSampling ? 1 : readOnlyGlobals.pktSampleRate;
  struct pcap_pkthdr h;
  int rc;
  unsigned long thread_id = (unsigned long)_thid;
  u_int num_read_pkts = 0;

  traceEvent(TRACE_INFO, "Fetch packets thread started [thread %lu]", thread_id);

  if(readOnlyGlobals.fakePacketCapture) {
    u_int8_t a = 0, b = 0;
    u_char pkt[] = {
      0x0, 0x1, 0x2, 0x3, 0x4, 0x5,
      0x6, 0x7, 0x8, 0x9, 0x0a, 0x0b,
      0x08, 0x00,
      0x45, 0x0,
      0x0, 0xcc, 0x95, 0x5, 0x40, 0x00, 0x40, 0x06,
      0x6d, 0x21, 0xc0, 0xa8, 0x01, 0x21, 0x4a, 0x7d,
      0x0, 0xc1, 0x00, 0x50, 0x3d, 0xe1,
      0xf0, 0xff, 0x3a, 0xf0, 0xdd, 0x0b, 0x80, 0x18,
      0x00, 0x0<5, 0x38, 0x6d, 0x00, 0x00, 0x01, 0x01,
      0x08, 0x0a, 0x00, 0x09, 0x65, 0xac, 0x98, 0xb7,
      0xc9, 0x3f
    };

    h.len = h.caplen = sizeof(pkt);
    readOnlyGlobals.datalink = DLT_EN10MB;

    while(!readWriteGlobals->shutdownInProgress) {
      h.ts.tv_sec = time(NULL);

      pkt[29] = a++, pkt[30] = b;
      decodePacket(thread_id,
		   -1 /* input interface id */,
		   &h, pkt,
		   readOnlyGlobals.fakePktSampling, readOnlyGlobals.pktSampleRate,
		   NO_INTERFACE_INDEX, NO_INTERFACE_INDEX,
		   0 /* Unknown sender */, 0 /* packet hash */);

      /* Rotating source IP */
      if(a == 0)
	b++;
    }
  } else {
#if 0
    setThreadAffinity(thread_id % readOnlyGlobals.numProcessThreads);
#endif

    packet = (char*)malloc(readOnlyGlobals.snaplen+1);
    if(packet == NULL) {
      traceEvent(TRACE_WARNING, "Not enough memory: fetchPackets(%d) leaving", thread_id);
      return(NULL);
    }

    while(!readWriteGlobals->shutdownInProgress) {
      /* traceEvent(TRACE_INFO, "fetchPackets(%d)", (int)notUsed); */
      if(readOnlyGlobals.fakePktSampling || (readOnlyGlobals.pktSampleRate == 1)) {
#ifdef DEBUG
	traceEvent(TRACE_INFO, "Processing packet");
#endif

	rc = next_pcap_packet(readOnlyGlobals.pcapPtr, &h, packet);

#if 0
	if((rc == 0) && (h->caplen == 0)) rc = -2; /* Sanity check */
#endif

	if((rc > 0) && (packet != NULL))
	  decodePacket(thread_id,
		       -1 /* input interface id */,
		       &h, packet,
		       readOnlyGlobals.fakePktSampling, readOnlyGlobals.pktSampleRate,
		       NO_INTERFACE_INDEX, NO_INTERFACE_INDEX,
		       0 /* Unknown sender */, 0 /* packet hash */);

	idleThreadTask(thread_id);
      } else {
	if(packetToGo > 1) {
	  rc = next_pcap_packet(readOnlyGlobals.pcapPtr, &h, packet);

	  if((rc == 1) && (packet != NULL)) {
	    packetToGo--;
#ifdef DEBUG
	    traceEvent(TRACE_INFO, "Discarded packet [%d packets to go]", packetToGo-1);
#endif
	  } else if(rc == -2) {
	    break; /* Captured file is over */
	  }
	  continue;
	} else {
#ifdef DEBUG
	  traceEvent(TRACE_INFO, "Processing packet");
#endif

	  rc = next_pcap_packet(readOnlyGlobals.pcapPtr, &h, packet);

	  if((rc == 0) && (h.caplen == 0)) rc = -2; /* Sanity check */
	  if((rc >= 0) && (packet != NULL)) {
	    decodePacket(thread_id,
			 -1 /* input interface id */,
			 &h, packet,
			 readOnlyGlobals.fakePktSampling, readOnlyGlobals.pktSampleRate,
			 NO_INTERFACE_INDEX, NO_INTERFACE_INDEX,
			 0 /* Unknown sender */, 0 /* packet hash */);
	    packetToGo = readOnlyGlobals.fakePktSampling ? 1 : readOnlyGlobals.pktSampleRate;
	  }
	}
      }

      if(rc < 0) {
	if(rc == -2) {
	  /* Captured file is over */
	  break;
	} else if(rc == -1) {
#ifdef WIN32
	  break;
#endif
	  if(!readWriteGlobals->shutdownInProgress) {
	    traceEvent(TRACE_ERROR, "Error while reading packets: '%s'",
		       pcap_geterr(readOnlyGlobals.pcapPtr));
	    term_pcap(&readOnlyGlobals.pcapPtr);
	    readOnlyGlobals.pcapPtr = NULL;
	    rc = restoreInterface(ebuf);
	    if(rc < 0) break;
	  }
	}
      } else if(rc == 0) {
	/* No more packets to read if reading from file */
	if(readOnlyGlobals.pcapFile != NULL)
	  break;
      }

      if(readOnlyGlobals.capture_num_packet_and_quit > 1)
	readOnlyGlobals.capture_num_packet_and_quit--;
      else if(readOnlyGlobals.capture_num_packet_and_quit == 1)
	readWriteGlobals->shutdownInProgress = 1;
    } /* while */

    readWriteGlobals->numTerminatedFetchPackets++;

    free((char*)packet);
  }

  return(NULL);
}

/* ****************************************************** */

void init_globals(void) {
  memset(&readOnlyGlobals, 0, sizeof(readOnlyGlobals));

  readWriteGlobals = (ReadWriteGlobals*)calloc(1, sizeof(ReadWriteGlobals));
  if(!readWriteGlobals) {
    traceEvent(TRACE_ERROR, "Not enough memory");
    exit(-1);
  }

  memset(&readOnlyGlobals, 0, sizeof(readOnlyGlobals));
  readOnlyGlobals.tunnel_mode = 0;
  readOnlyGlobals.promisc_mode = 1;
  readOnlyGlobals.maxNumActiveFlows = (u_int)-1;
  readOnlyGlobals.idTemplate = DEFAULT_TEMPLATE_ID;

#ifdef HAVE_FASTBIT
  readOnlyGlobals.fastbit_mins_rotation = DEFAULT_FASTBIT_MINS_ROTATION;
#endif
#ifdef linux
  readOnlyGlobals.cpuAffinity = NULL; /* no affinity */
#endif
  readOnlyGlobals.scanCycle = 30 /* sec */;
  readOnlyGlobals.handleFragments = 1;
  readOnlyGlobals.inputInterfaceIndex = DEFAULT_INPUT_INTERFACE_INDEX;
  readOnlyGlobals.outputInterfaceIndex = DEFAULT_OUTPUT_INTERFACE_INDEX;
  readOnlyGlobals.file_dump_timeout = 60;
  readOnlyGlobals.templatePacketsDelta = TEMPLATE_PACKETS_DELTA;

  /* Resever one core as a thread is used for packet dequeueing */
  readOnlyGlobals.numProcessThreads = 1;

  readOnlyGlobals.enableHostStats = 0;
  readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].templateBufMax = readOnlyGlobals.templateBuffers[V6_TEMPLATE_INDEX].templateBufMax = NETFLOW_MAX_BUFFER_LEN;
  readOnlyGlobals.optionTemplateBufMax = NETFLOW_MAX_BUFFER_LEN;
  readOnlyGlobals.dumpFormat = text_format;
  readOnlyGlobals.traceLevel = 2;
  readOnlyGlobals.idleTimeout = DUMP_TIMEOUT;
  readOnlyGlobals.deferredHostUpdate = 0; // FIX
  readOnlyGlobals.lifetimeTimeout = 4*DUMP_TIMEOUT;
  readOnlyGlobals.sendTimeout = DUMP_TIMEOUT;
  readWriteGlobals->lastMaxBucketSearch = 5; /* Don't bother with values < 5 */
  readOnlyGlobals.pcapPtr = NULL;
  readOnlyGlobals.csv_separator = strdup("|");
  readOnlyGlobals.nfLitePluginEnabled = 0;

#ifdef DEMO_MODE
  readOnlyGlobals.demo_mode = 1;
#endif
}

/* ****************************************************** */

static void printCopyrights(void) {
#ifdef HAVE_GEOIP
  if(readOnlyGlobals.geo_ip_city_db != NULL)
    traceEvent(TRACE_NORMAL, "%s", GeoIP_database_info(readOnlyGlobals.geo_ip_city_db));
  if(readOnlyGlobals.geo_ip_asn_db != NULL)
    traceEvent(TRACE_NORMAL, "%s", GeoIP_database_info(readOnlyGlobals.geo_ip_asn_db));
#endif
}

/* ****************************************************** */

static void compileTemplates(u_int8_t reloadTemplate) {
  u_int num_runs, i;
  const u_int templatedBufferLen = 4096;
  char *baseTempleteBufferV4, *baseTempleteBufferV6, *stringBuffer;

  if((baseTempleteBufferV4 = calloc(1, templatedBufferLen)) == NULL) {
    traceEvent(TRACE_ERROR, "Not enough memory?");
    return;
  }

  if((baseTempleteBufferV6 = calloc(1, templatedBufferLen)) == NULL) {
    traceEvent(TRACE_ERROR, "Not enough memory?");
    return;
  }

  if((stringBuffer = calloc(1, templatedBufferLen)) == NULL) {
    traceEvent(TRACE_ERROR, "Not enough memory?");
    return;
  }

  traceEvent(TRACE_INFO, "Compiling flow templates...");

  if(reloadTemplate)
    stopCaptureFlushAll();

  if(readOnlyGlobals.netFlowVersion == 5) {
    readOnlyGlobals.stringTemplateV4 = strdup(DEFAULT_V9_IPV4_TEMPLATE);
    compileTemplate(readOnlyGlobals.stringTemplateV4,
		    readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].v9TemplateElementList,
		    TEMPLATE_LIST_LEN, 0, 0 /* IPv4 */);

    compileTemplate(readOnlyGlobals.stringTemplateV4,
		    readOnlyGlobals.userTemplateBuffer.v9TemplateElementList,
		    TEMPLATE_LIST_LEN, 0, 0 /* IPv4 */);

    if(!reloadTemplate) {
#ifdef HAVE_MYSQL
      init_db_table();
#endif

#ifdef HAVE_FASTBIT
      init_fastbit(readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].v9TemplateElementList,
		   NULL /* No configuration file */);
#endif
    }

    buildActivePluginsList(readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].v9TemplateElementList);
  } else if(readOnlyGlobals.netFlowVersion == 9 || readOnlyGlobals.netFlowVersion == 10) {
    u_int flowLen;
    char *elem;

    if(readOnlyGlobals.baseTemplateBufferV4 == NULL) {
      traceEvent(TRACE_WARNING, "You selected v9/IPFIX without specifying a template (-T).");
      traceEvent(TRACE_WARNING, "The default template will be used");
      readOnlyGlobals.baseTemplateBufferV4 = strdup((readOnlyGlobals.netFlowVersion == 9) ?
						    DEFAULT_V9_IPV4_TEMPLATE : DEFAULT_IPFIX_IPV4_TEMPLATE);
    }
    
    readOnlyGlobals.stringTemplateV4 = strdup(readOnlyGlobals.baseTemplateBufferV4);
    if(strchr(readOnlyGlobals.stringTemplateV4, '%') == NULL)
      traceEvent(TRACE_WARNING, "The template does not contain any '%%': please check its format");

    if(strstr(readOnlyGlobals.stringTemplateV4, "%FLOW_PROTO_PORT") != NULL)
      loadApplProtocols();

    if((strstr(readOnlyGlobals.stringTemplateV4, "%IN_BYTES")
	&& strstr(readOnlyGlobals.stringTemplateV4, "%OUT_BYTES"))
       || (strstr(readOnlyGlobals.stringTemplateV4, "%IN_PKTS")
	   && strstr(readOnlyGlobals.stringTemplateV4, "%OUT_PKTS")))
      readOnlyGlobals.dontSentBidirectionalV9Flows = 1;
    else
      readOnlyGlobals.dontSentBidirectionalV9Flows = 0;

    if(strstr(readOnlyGlobals.stringTemplateV4, "%RETRANSMITTED_")
       || strstr(readOnlyGlobals.stringTemplateV4, "%OOORDER")
       || strstr(readOnlyGlobals.stringTemplateV4, "_NW_DELAY_")
       )
      readOnlyGlobals.enableTcpSeqStats = 1, readOnlyGlobals.enableExtBucket = 1,
	readOnlyGlobals.ignoreTos = 1; /* Unless we ignore it, we cannot properly compute bi-directional stats */

    if(strstr(readOnlyGlobals.stringTemplateV4, "%NUM_PKTS_")
       || strstr(readOnlyGlobals.stringTemplateV4, "%PACKET_VARIANCE")
       )
      readOnlyGlobals.enablePacketStats = 1, readOnlyGlobals.enableExtBucket = 1;

    if(strstr(readOnlyGlobals.stringTemplateV4, "%MPLS"))
      readOnlyGlobals.enableExtBucket = 1;

    if(strstr(readOnlyGlobals.stringTemplateV4, "%APPL_LATENCY"))
      readOnlyGlobals.enableLatencyStats = 1, readOnlyGlobals.enableExtBucket = 1;

    if(strstr(readOnlyGlobals.stringTemplateV4, "%JITTER") != NULL)
      readOnlyGlobals.calculateJitter = 1;

    if(strstr(readOnlyGlobals.stringTemplateV4, "%L7_PROTO") != NULL)
      readOnlyGlobals.enable_l7_protocol_discovery = 1;
    else
      readOnlyGlobals.l7.discard_unknown_flows = 0;

    switch(readOnlyGlobals.l7.discard_unknown_flows) {
    case 1:
      traceEvent(TRACE_NORMAL, "Flows with unknown L7 protocols will be discarded");
      break;
    case 2:
      traceEvent(TRACE_NORMAL, "Flows with known L7 protocols will be discarded");
      break;
    }

#ifdef HAVE_GEOIP
    if(strstr(readOnlyGlobals.stringTemplateV4, "_COUNTRY")
       || strstr(readOnlyGlobals.stringTemplateV4, "_CITY"))
      readOnlyGlobals.enableGeoIP = 1;
#endif

    /*
      Optimization for NetFlow v9: discard fields that are not needed
    */
    if((strstr(readOnlyGlobals.stringTemplateV4, "%IPV4_SRC_ADDR") == NULL)
       && (strstr(readOnlyGlobals.stringTemplateV4, "%IPV4_DST_ADDR") == NULL)
       && (strstr(readOnlyGlobals.stringTemplateV4, "%IPV6_SRC_ADDR") == NULL)
       && (strstr(readOnlyGlobals.stringTemplateV4, "%IPV6_DST_ADDR") == NULL)
       ) {
      traceEvent(TRACE_WARNING,
		 "IPv4/v6 addresses will be ignored (your template lacks %%IPV4_XXX_ADDR/%%IPV6_XXX_ADDR)");
      readOnlyGlobals.ignoreIP = 1;
    }

    if((strstr(readOnlyGlobals.stringTemplateV4, "%L4_SRC_PORT") == NULL)
       && (strstr(readOnlyGlobals.stringTemplateV4, "%L4_DST_PORT") == NULL)) {
      traceEvent(TRACE_WARNING,
		 "L4 ports will be ignored (your template lacks %%L4_SRC_PORT/%%L4_DST_PORT)");
      readOnlyGlobals.ignorePorts = readOnlyGlobals.ignorePorts = 1;
    }

    if(strstr(readOnlyGlobals.stringTemplateV4, "%PROTOCOL") == NULL) {
      traceEvent(TRACE_WARNING,
		 "Protocol will be ignored (your template lacks %%PROTOCOL)");
      readOnlyGlobals.ignoreProtocol = 1;
    }

    if(readOnlyGlobals.netFlowVersion == 5) {
      readOnlyGlobals.enableExtBucket = 0;

      if(strstr(readOnlyGlobals.stringTemplateV4, "%ICMP_TYPE"))
	readOnlyGlobals.usePortsForICMP = 0;
      else {
	readOnlyGlobals.usePortsForICMP = 1;
	traceEvent(TRACE_INFO,
		   "TCP/UDP port will carry ICMP type/code information (your template lacks %%ICMP_TYPE)");
      }
    }

    /* Original user-specified template */
    readOnlyGlobals.userStringTemplate = strdup(readOnlyGlobals.baseTemplateBufferV4);
    compileTemplate(readOnlyGlobals.userStringTemplate,
		    readOnlyGlobals.userTemplateBuffer.v9TemplateElementList,
		    TEMPLATE_LIST_LEN, 0, 0 /* IPv4/v6 */);

    /* This is used to enable plugins and set in use to template elements */
    sanitizeV4Template(readOnlyGlobals.stringTemplateV4);
    compileTemplate(readOnlyGlobals.stringTemplateV4,
		    readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].v9TemplateElementList,
		    TEMPLATE_LIST_LEN, 0, 0 /* IPv4/v6 */);

    /* Dummy code for enabling V6 support */
    v4toV6Template(readOnlyGlobals.stringTemplateV4);
    compileTemplate(readOnlyGlobals.stringTemplateV4,
		    readOnlyGlobals.templateBuffers[V6_TEMPLATE_INDEX].v9TemplateElementList,
		    TEMPLATE_LIST_LEN, 0, 0 /* IPv4/v6 */);

    i = 0;
    while(ver9_templates[i].netflowElementName != NULL) {
      if(ver9_templates[i].isInUse) {
	switch(ver9_templates[i].protoMode) {
	case BOTH_IPV4_IPV6:
	  snprintf(&baseTempleteBufferV4[strlen(baseTempleteBufferV4)],
		   templatedBufferLen-strlen(baseTempleteBufferV4)," %%%s",
		   ver9_templates[i].netflowElementName);

	  snprintf(&baseTempleteBufferV6[strlen(baseTempleteBufferV6)],
		   templatedBufferLen-strlen(baseTempleteBufferV6)," %%%s",
		   ver9_templates[i].netflowElementName);
	  break;
	case ONLY_IPV4:
	  snprintf(&baseTempleteBufferV4[strlen(baseTempleteBufferV4)],
		   templatedBufferLen-strlen(baseTempleteBufferV4)," %%%s",
		   ver9_templates[i].netflowElementName);
	  break;
	case ONLY_IPV6:
	  snprintf(&baseTempleteBufferV6[strlen(baseTempleteBufferV6)],
		   templatedBufferLen-strlen(baseTempleteBufferV6)," %%%s",
		   ver9_templates[i].netflowElementName);
	  break;
	}
      }

      i++;
    }

    /*
      We need to check if the user has
      1. created a template for both IPv4 and IPv6: in this case we will not change
         the template. This is useful for those situations where both IPv4 and v6 will
	 be mixed into the same packet (e.g. IPv4 tunnel of IPv6 traffic)
      2. if we have a single IPv4-only template we will create a similar IPv6 template
    */

    /*
       Dummy calls for enabling all plugins: their values will be overwritten
       but they are functional to nProbe
    */
    compileTemplate(readOnlyGlobals.stringTemplateV4,
		    readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].v9TemplateElementList,
		    TEMPLATE_LIST_LEN, 0, 0 /* IPv4/v6 */);
    buildActivePluginsList(readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].v9TemplateElementList);

    /* ******** */

    free(readOnlyGlobals.stringTemplateV4); readOnlyGlobals.stringTemplateV4 = strdup(baseTempleteBufferV4);
    compileTemplate(readOnlyGlobals.stringTemplateV4,
		    readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].v9TemplateElementList,
		    TEMPLATE_LIST_LEN, 0, 0 /* IPv4/v6 */);
    readOnlyGlobals.stringTemplateV6 = NULL;

    flowPrintf(readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].v9TemplateElementList,
	       1 /* IPv4 */, readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].templateBuffer,
	       &readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].templateBufBegin,
	       &readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].templateBufMax,
	       &readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].numTemplateFieldElements,
	       1, NULL, 0, 0, 0);
    readOnlyGlobals.numActiveTemplates++;

    /* ******* */

    /* Do the check below before we deallocate the string */
    readOnlyGlobals.stringTemplateV6 = strdup(baseTempleteBufferV6);
    compileTemplate(readOnlyGlobals.stringTemplateV6,
		    readOnlyGlobals.templateBuffers[V6_TEMPLATE_INDEX].v9TemplateElementList,
		    TEMPLATE_LIST_LEN, 0, 1 /* IPv6 ONLY template */);
    flowPrintf(readOnlyGlobals.templateBuffers[V6_TEMPLATE_INDEX].v9TemplateElementList,
	       0 /* IPv6 */, readOnlyGlobals.templateBuffers[V6_TEMPLATE_INDEX].templateBuffer,
	       &readOnlyGlobals.templateBuffers[V6_TEMPLATE_INDEX].templateBufBegin,
	       &readOnlyGlobals.templateBuffers[V6_TEMPLATE_INDEX].templateBufMax,
	       &readOnlyGlobals.templateBuffers[V6_TEMPLATE_INDEX].numTemplateFieldElements,
	       1, NULL, 0, 0, 0);
    readOnlyGlobals.numActiveTemplates++;

    /* ******** */

    for(i=0; i<readOnlyGlobals.num_active_plugins; i++) {
      int j;
      V9V10TemplateElementId *el = readOnlyGlobals.all_active_plugins[i]->pluginFlowConf();

      strcpy(stringBuffer, baseTempleteBufferV4);

      j = 0;
      while(el[j].netflowElementName != NULL) {
	if(el[j].isInUse && (el[j].protoMode != ONLY_IPV6))
	  snprintf(&stringBuffer[strlen(stringBuffer)],
		   templatedBufferLen-strlen(stringBuffer),
		   " %%%s", el[j].netflowElementName);

	j++;
      }

      compileTemplate(stringBuffer,
		      readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].v9TemplateElementList,
		      TEMPLATE_LIST_LEN, 0, 0 /* IPv4 */);
      readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].templateBufMax = NETFLOW_MAX_BUFFER_LEN;
      flowPrintf(readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].v9TemplateElementList,
		 1 /* IPv4 */, readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].templateBuffer,
		 &readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].templateBufBegin,
		 &readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].templateBufMax,
		 &readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].numTemplateFieldElements,
		 1, NULL, 0, 0, 0);

      readOnlyGlobals.all_active_plugins[i]->v4TemplateIdx = readOnlyGlobals.numActiveTemplates;

      readOnlyGlobals.numActiveTemplates++;

      if(readOnlyGlobals.stringTemplateV6) {
	strcpy(stringBuffer, baseTempleteBufferV6);

	j = 0;
	while(el[j].netflowElementName != NULL) {
	  if(el[j].isInUse && (el[j].protoMode != ONLY_IPV4))
	    snprintf(&stringBuffer[strlen(stringBuffer)],
		     templatedBufferLen-strlen(stringBuffer),
		     " %%%s", el[j].netflowElementName);

	  j++;
	}

	compileTemplate(stringBuffer,
			readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].v9TemplateElementList,
			TEMPLATE_LIST_LEN, 0, 1 /* IPv6 */);
	readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].templateBufMax = NETFLOW_MAX_BUFFER_LEN;
	flowPrintf(readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].v9TemplateElementList,
		   0 /* IPv6 */, readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].templateBuffer,
		   &readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].templateBufBegin,
		   &readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].templateBufMax,
		   &readOnlyGlobals.templateBuffers[readOnlyGlobals.numActiveTemplates].numTemplateFieldElements,
		   1, NULL, 0, 0, 0);
	readOnlyGlobals.all_active_plugins[i]->v6TemplateIdx = readOnlyGlobals.numActiveTemplates;
	readOnlyGlobals.numActiveTemplates++;
      }
    } /* for */

    /* Option template */
    compileTemplate(V9_OPTION_TEMPLATE, readOnlyGlobals.v9OptionTemplateElementList,
		    TEMPLATE_LIST_LEN, 1, 0 /* IPv4/v6 */);
    flowPrintf(readOnlyGlobals.v9OptionTemplateElementList,
	       1 /* IPv4 */, readOnlyGlobals.optionTemplateBuffer,
	       &readOnlyGlobals.optionTemplateBufBegin, &readOnlyGlobals.optionTemplateBufMax,
	       &readOnlyGlobals.numOptionTemplateFieldElements, 1, NULL, 0, 0, 1);

    flowLen = 0;
    if(readOnlyGlobals.traceMode == 2) traceEvent(TRACE_INFO, "Scanning flow template...");

    for(num_runs = 0; num_runs < readOnlyGlobals.numActiveTemplates; num_runs++) {
      V9V10TemplateElementId **elems;
      u_int tot, elId;

      elems = readOnlyGlobals.templateBuffers[num_runs].v9TemplateElementList;
      if(elems[0] == NULL) continue;

      if(readOnlyGlobals.traceMode == 2)
	traceEvent(TRACE_INFO, "Template [id=%u]", readOnlyGlobals.idTemplate + num_runs);

      for(i=0, tot = 0, elId = 0; i<TEMPLATE_LIST_LEN; i++) {
	if(elems[i] != NULL) {
	  tot += elems[i]->templateElementLen;
	  if(readOnlyGlobals.traceMode == 2)
	    traceEvent(TRACE_INFO, "Found %20s [num %d][id %d][%d bytes][total %d bytes]",
		       elems[i]->netflowElementName, ++elId,
		       (elems[i]->templateElementEnterpriseId == NTOP_ENTERPRISE_ID)
		       ? elems[i]->templateElementId-NTOP_BASE_ID : elems[i]->templateElementId,
		       elems[i]->templateElementLen, tot);
	} else
	  break;

	if(tot > flowLen) flowLen = tot;
      }
    }

    if((readOnlyGlobals.userTemplateBuffer.v9TemplateElementList[0] == NULL)
       && (readOnlyGlobals.numCollectors == 0)
       && readOnlyGlobals.none_specified) {
      traceEvent(TRACE_ERROR, "-D is mandatory when '-n none' is specified");
      exit(0);
    }

#ifdef HAVE_MYSQL
    init_db_table();
#endif

#ifdef HAVE_FASTBIT
    init_fastbit(readOnlyGlobals.templateBuffers[V4_TEMPLATE_INDEX].v9TemplateElementList, NULL /* No configuration file */);
#endif

    if(flowLen > 0) {
      if(readOnlyGlobals.traceMode == 2) traceEvent(TRACE_INFO, "Scanning option template...");
      for(i=0; i<readOnlyGlobals.numOptionTemplateFieldElements; i++) {
	if(readOnlyGlobals.v9OptionTemplateElementList[i] != NULL) {
	  readOnlyGlobals.optionTemplateFlowSize += readOnlyGlobals.v9OptionTemplateElementList[i]->templateElementLen;

	  if(readOnlyGlobals.traceMode == 2) {
	    traceEvent(TRACE_INFO, "Found %20s [id %d][%u bytes][total %d bytes]",
		       readOnlyGlobals.v9OptionTemplateElementList[i]->netflowElementName,
		       readOnlyGlobals.v9OptionTemplateElementList[i]->templateElementId,
		       readOnlyGlobals.v9OptionTemplateElementList[i]->templateElementLen,
		       (int)readOnlyGlobals.optionTemplateFlowSize);
	  }
	} else
	  break;
      }

      readOnlyGlobals.templateFlowSize = 8;

      for(i=0; i<readOnlyGlobals.numActiveTemplates; i++)
	readOnlyGlobals.templateFlowSize += readOnlyGlobals.templateBuffers[i].templateBufBegin;

      readOnlyGlobals.templateFlowSize += ((12 + readOnlyGlobals.optionTemplateBufBegin)
					   + (4 + readOnlyGlobals.optionTemplateFlowSize)
					   + (flowLen - 1) /* Avoid rounding problems */
					    );

      readOnlyGlobals.templateFlowSize /= flowLen;

      if(readOnlyGlobals.minNumFlowsPerPacket == (u_short)-1) {
	/*
	  As with NetFlow v5, we suppose that a UDP packet can fit up to 1440
	  bytes (alias NETFLOW_MAX_BUFFER_LEN) of payload for NetFlow flows.
	*/
	readOnlyGlobals.minNumFlowsPerPacket = (NETFLOW_MAX_BUFFER_LEN/flowLen)-1;
	traceEvent(TRACE_INFO, "Each flow is %d bytes long", flowLen);
	traceEvent(TRACE_INFO, "The # packets per flow has been set to %d",
		   readOnlyGlobals.minNumFlowsPerPacket);
      } else {
	if((readOnlyGlobals.minNumFlowsPerPacket*flowLen) >= NETFLOW_MAX_BUFFER_LEN) {
	  traceEvent(TRACE_WARNING,
		     "Too many flows (%d) per packet specified using -m.",
		     readOnlyGlobals.minNumFlowsPerPacket);
	  readOnlyGlobals.minNumFlowsPerPacket = (NETFLOW_MAX_BUFFER_LEN/flowLen)-1;
	  traceEvent(TRACE_INFO, "The # packets per flow has been set to %d",
		     readOnlyGlobals.minNumFlowsPerPacket);
	}
      }
    } else {
      readOnlyGlobals.netFlowVersion = 5;
      traceEvent(TRACE_INFO, "The flow size is zero. Switching back to v5");
    }
  }


#ifdef HAVE_GEOIP
  if(readOnlyGlobals.geo_ip_asn_db == NULL)
    readASs("GeoIPASNum.dat");

  if(!readOnlyGlobals.enableGeoIP) {
    if(readOnlyGlobals.geo_ip_city_db != NULL) {
      GeoIP_delete(readOnlyGlobals.geo_ip_city_db);
      readOnlyGlobals.geo_ip_city_db = NULL;
    }
  } else {
    if(readOnlyGlobals.geo_ip_city_db == NULL)
      readCities("GeoLiteCity.dat");
  }
#endif

  /* Allocate memory for template buffers */
  for(i=0; i<readOnlyGlobals.numActiveTemplates; i++) {
    if((readOnlyGlobals.templateBuffers[i].buffer = (char*)malloc(NETFLOW_MAX_BUFFER_LEN)) == NULL) {
      traceEvent(TRACE_ERROR, "Not enough memory ?");
      exit(0); /* If we don't have enough memory now, we better quit */
    }
  }

  if(reloadTemplate) {
    readWriteGlobals->stopPacketCapture = 0;
    traceEvent(TRACE_INFO, "nProbe is now operational...");
  }

  free(baseTempleteBufferV4);
  free(baseTempleteBufferV6);
  free(stringBuffer);
}

/* ****************************************************** */

static void debug_printf(u32 protocol, void *id_struct, ipq_log_level_t log_level, const char *format, ...) { ; }
static void *malloc_wrapper(unsigned long size) { return malloc(size); }

static void initL7Discovery(void) {
  IPOQUE_PROTOCOL_BITMASK all;
  u32 detection_tick_resolution = 1000;

  if(!readOnlyGlobals.enable_l7_protocol_discovery) return;

  readOnlyGlobals.l7.l7handler = ipoque_init_detection_module(detection_tick_resolution,
							      malloc_wrapper, debug_printf);
  if(readOnlyGlobals.l7.l7handler == NULL) {
    traceEvent(TRACE_ERROR, "Unable to initialize L7 engine: disabling L7 discovery");
    readOnlyGlobals.enable_l7_protocol_discovery = 0;
    return;
  }

  // enable all protocols
  IPOQUE_BITMASK_SET_ALL(all);
  ipoque_set_protocol_detection_bitmask2(readOnlyGlobals.l7.l7handler, &all);

  readOnlyGlobals.l7.proto_size = ipoque_detection_get_sizeof_ipoque_id_struct();
  readOnlyGlobals.l7.flow_struct_size = ipoque_detection_get_sizeof_ipoque_flow_struct();
}

/* ****************************************************** */

int
#ifdef WIN32
nprobe_main
#else
main
#endif
(int argc, char *argv[]) {
  char ebuf[PCAP_ERRBUF_SIZE] = { '\0' }, *sysId;
  u_int i, idx;

#ifdef WIN32
  int optind = 0;

  ptw32_processInitialize();
#endif


#if defined(HAVE_LICENSE) || defined(USE_SPARROW) || defined(WIN32)

  /* NOTE
     As the getopt is manipulated this MUST be the
     first function to be called at startup
  */
  if(!readOnlyGlobals.demo_mode) {
    extern int optind, opterr, optopt;
    extern int verify_application_instances(char *application_name, char *out_buf, int out_buf_len);
    int t_optind, t_opterr, t_optopt, num_instances;
    char out_buf[8192];

    /* save values of optind, opterr and optopt because license library
     * Calls getopt_long
     */
    t_optind=optind;  t_opterr=opterr;  t_optopt=optopt;

    if((argc == 2) && (!strcmp(argv[1], "-v"))) {
      probeVersion();
      exit(0);
    } else if((argc == 2) && (!strcmp(argv[1], "--nprobe-version"))) {
      printf("%s\n", version);
      exit(0);
    } else if((argc == 2) && (!strcmp(argv[1], "-h"))) {
      init_globals();
      usage();
      exit(0);
    }

    if(
#ifdef USE_SPARROW
  (checkSparrow() == 0) /* No/bad license: switching to demo mode */
#else
		(!isGoodLicense())
#endif
		) {
#ifndef USE_SPARROW
      char *sysId;
#endif

      traceEvent(TRACE_ERROR, "Invalid or missing nProbe license"
#ifndef USE_SPARROW
		  " ("
#ifndef WIN32
		 "/etc/"
#endif
		 LICENSE_FILE_NAME")"
#endif
		 );
#ifndef USE_SPARROW
	  traceEvent(TRACE_ERROR, "for %s", (sysId = getSystemId()));
      /* exit(0); */
      free(sysId);
#endif
      readOnlyGlobals.demo_mode = 1;
    } else {
      traceEvent(TRACE_NORMAL, "Valid nProbe license found");

      if((num_instances = verify_application_instances("nprobe", out_buf, sizeof(out_buf))) != 0) {
	traceEvent(TRACE_ERROR, "Too many nProbe instances (%d) running", num_instances);
	/* exit(-1); */
	readOnlyGlobals.demo_mode = 1;
      }
    }

    if(readOnlyGlobals.demo_mode) {
      traceEvent(TRACE_ERROR, "***************************************************");
      traceEvent(TRACE_ERROR, "**                                               **");
      traceEvent(TRACE_ERROR, "**  Switching to DEMO MODE due to license error  **");
      traceEvent(TRACE_ERROR, "**                                               **");
      traceEvent(TRACE_ERROR, "**  Create your nProbe license at                **");
      traceEvent(TRACE_ERROR, "**       http://www.nmon.net/mklicense/          **");
      traceEvent(TRACE_ERROR, "**                                               **");
      traceEvent(TRACE_ERROR, "***************************************************");
    }

    /* restore values */
    optind=t_optind;  opterr=t_opterr;  optopt=t_optopt;
  }
#endif

  init_globals();

  if(readOnlyGlobals.demo_mode) {
    printf("*************************************************************\n");
    printf("* NOTE: This is a DEMO version limited to %d flows export. *\n",
	   MAX_DEMO_FLOWS);
    printf("*************************************************************\n\n");
  }

#ifdef WIN32
  initWinsock32();
#else
  setprotoent(1); setservent(1); /* Improve protocol/port lookup performance */
#endif

  argc_ = argc;
  argv_ = (char**)argv;
  if(parseOptions(argc, argv, 0) == -1) exit(0);

  // readOnlyGlobals.traceMode = 2, traceLevel = 5; // FIX
  initPlugins(argc, argv);

  traceEvent(TRACE_NORMAL, "Welcome to nprobe v.%s for %s", version, osName);
  printCopyrights();

#if defined(HAVE_LICENSE) || defined(WIN32)
  traceEvent(TRACE_NORMAL, "SystemID: %s", (sysId = getSystemId()));
  free(sysId);
#endif

#ifndef WIN32
  if(readOnlyGlobals.useSyslog)
    openlog(readOnlyGlobals.nprobeId, LOG_PID ,LOG_DAEMON);
#endif

  memset(&readWriteGlobals->theFlowHash, 0, sizeof(readWriteGlobals->theFlowHash));
  readWriteGlobals->shutdownInProgress = 0;
  readWriteGlobals->flowExportStats.totExportedBytes = 0;
  readWriteGlobals->flowExportStats.totExportedPkts = readWriteGlobals->flowExportStats.totExportedFlows = 0;

  createCondvar(&readWriteGlobals->exportQueueCondvar);
  createCondvar(&readWriteGlobals->termCondvar);
  pthread_rwlock_init(&readWriteGlobals->exportMutex, NULL);

  for(i=0; i<NUM_FRAGMENT_LISTS; i++)
    pthread_rwlock_init(&readWriteGlobals->fragmentMutex[i], NULL);

#ifdef HAVE_GEOIP
  pthread_rwlock_init(&readWriteGlobals->geoipRwLock, NULL);
#endif

  pthread_rwlock_init(&readOnlyGlobals.ticksLock, NULL);
  pthread_rwlock_init(&readWriteGlobals->exportRwLock, NULL);
  pthread_rwlock_init(&readWriteGlobals->rwGlobalsRwLock, NULL);
  pthread_rwlock_init(&readWriteGlobals->collectorRwLock, NULL);
  pthread_rwlock_init(&readWriteGlobals->collectorCounterLock, NULL);
  pthread_rwlock_init(&readWriteGlobals->pcapLock, NULL);
  pthread_rwlock_init(&readWriteGlobals->checkExportLock, NULL);

  for(i=0; i<MAX_HASH_MUTEXES; i++)
    pthread_mutex_init(&readWriteGlobals->hostHashMutex[i], NULL);

  /* FIX
     if(textFormat
     && ((strstr(textFormat, "%js") != NULL)
     || (strstr(textFormat, "%jd") != NULL)))
     calculateJitter = 1;
  */

  if(readOnlyGlobals.bidirectionalFlows && (readOnlyGlobals.netFlowVersion == 5)) {
    traceEvent(TRACE_WARNING, "Bi-directional flows are not supported by NetFlowV5: disabled");
    readOnlyGlobals.bidirectionalFlows = 0;
  }

  compileTemplates(0);

  if((readOnlyGlobals.netFlowVersion != 5) && readOnlyGlobals.ignoreIP)
    traceEvent(TRACE_WARNING, "Your template ignores IP addresses: your collector might ignore these flows.");

  if((readOnlyGlobals.dirPath != NULL)
     && (readOnlyGlobals.userTemplateBuffer.v9TemplateElementList[0] == NULL)) {
    traceEvent(TRACE_WARNING,
	       "-P can be specified only with -D. Ignoring -P value [%s].",
	       readOnlyGlobals.dirPath);
    free(readOnlyGlobals.dirPath);
    readOnlyGlobals.dirPath = NULL;
  }

  if(readOnlyGlobals.ignorePorts && readOnlyGlobals.ignoreProtocol)
    readOnlyGlobals.handleFragments = 0;

#ifndef WIN32
  signal(SIGTERM, cleanup);
  signal(SIGINT,  cleanup);
  signal(SIGPIPE, brokenPipe);
  signal(SIGHUP,  reloadCLI);
#endif

  /* pcap-based sniffing */
  memset(readWriteGlobals->theFlowHash, 0, sizeof(readWriteGlobals->theFlowHash));

  if((readOnlyGlobals.collectorInPort == 0) || (readOnlyGlobals.captureDev != NULL)) {
    if((openDevice(ebuf, 1, (readOnlyGlobals.pcapFileList ? readOnlyGlobals.pcapFileList->path : NULL)) == -1)
       || ((readOnlyGlobals.pcapPtr == NULL)
	   && strcmp(readOnlyGlobals.captureDev, "none")
#ifdef HAVE_PF_RING
	   && (readWriteGlobals->ring == NULL)
	   && (readOnlyGlobals.metadata_fd == NULL)
#endif
	   )) {
      traceEvent(TRACE_ERROR, "Unable to open interface %s (%s)\n",
		 readOnlyGlobals.captureDev == NULL ? "<unknown>" : readOnlyGlobals.captureDev, ebuf);
      traceEvent(TRACE_ERROR, "Try using -i none if you do not want capture from a NIC");
      exit(-1);
    }

    if(readOnlyGlobals.pcapFileList != NULL) {
      struct fileList *next = readOnlyGlobals.pcapFileList->next;

      free(readOnlyGlobals.pcapFileList->path);
      free(readOnlyGlobals.pcapFileList);
      readOnlyGlobals.pcapFileList = next;
    }
  }

  if(readOnlyGlobals.numProcessThreads > 1) {
    u_int8_t have_pf_ring = 0;

    /*
       In order to support multithreaded packet processing
       we need to use PF_RING natively
    */
#ifdef HAVE_PF_RING
    if(readWriteGlobals->ring != NULL)
      have_pf_ring = 1;
#endif

    if(have_pf_ring) {
      /* We need to allocate per-thread packet queues */
      for(i=0; i<readOnlyGlobals.numProcessThreads; i++) {
	PacketQueue *queue = &readWriteGlobals->packetQueues[i];
	int j;

	createCondvar(&queue->dequeue_condvar);

	queue->queueSlots = (QueuedPacket*)calloc(sizeof(QueuedPacket), DEFAULT_QUEUE_CAPACITY);
	if(queue->queueSlots == NULL) {
	  traceEvent(TRACE_ERROR, "Not enough memory");
	  exit(-1);
	}

	for(j=0; j<DEFAULT_QUEUE_CAPACITY; j++) {
	  if((queue->queueSlots[j].p = malloc(readOnlyGlobals.snaplen)) == NULL) {
	    traceEvent(TRACE_ERROR, "Not enough memory");
	    exit(-1);
	  }
	}
      }
    } else {
      traceEvent(TRACE_WARNING, "Multithreaded processing is supported only with PF_RING");
      traceEvent(TRACE_WARNING, "Switching back to single thread processing");
      readOnlyGlobals.numProcessThreads = 1;
    }
  }

  memset(readWriteGlobals->flowHashRwLock, 0, sizeof(readWriteGlobals->flowHashRwLock));
  for(idx=0; idx<readOnlyGlobals.numProcessThreads; idx++) {
    for(i=0; i<MAX_HASH_MUTEXES; i++)
      pthread_rwlock_init(&readWriteGlobals->flowHashRwLock[idx][i], NULL);
  }

  for(idx=0; idx<readOnlyGlobals.numProcessThreads; idx++)
    allocateFlowHash(idx);

  for(i=0; i<readOnlyGlobals.numProcessThreads; i++) {
    readWriteGlobals->accumulateStats[i].pkts = 0, readWriteGlobals->accumulateStats[i].bytes = 0,
      readWriteGlobals->accumulateStats[i].tcpPkts = 0, readWriteGlobals->accumulateStats[i].tcpBytes = 0;
    readWriteGlobals->accumulateStats[i].udpPkts = 0, readWriteGlobals->accumulateStats[i].udpBytes = 0,
      readWriteGlobals->accumulateStats[i].icmpPkts = 0, readWriteGlobals->accumulateStats[i].icmpBytes = 0;
  }

  readWriteGlobals->lastSample = time(NULL);

  if(readOnlyGlobals.enableHttpPlugin
     || readOnlyGlobals.enableDnsPlugin)
    readOnlyGlobals.disableIPv6 = 0;

  if(readOnlyGlobals.disableIPv6) {
    traceEvent(TRACE_NORMAL, "IPv6 traffic will NOT be exported/accounted by this probe");
    traceEvent(TRACE_NORMAL, "due to configuration options (e.g. use NetFlow v9)");
  }

  traceEvent(TRACE_INFO, "The flows hash has %d buckets",
	     readOnlyGlobals.flowHashSize);
  traceEvent(TRACE_INFO, "Flows older than %d seconds will be exported",
	     readOnlyGlobals.lifetimeTimeout);

  traceEvent(TRACE_INFO, "Flows inactive for at least %d seconds will be exported",
	     readOnlyGlobals.idleTimeout);

  traceEvent(TRACE_INFO, "Expired flows will be checked every %d seconds",
	     readOnlyGlobals.scanCycle);
  traceEvent(TRACE_INFO, "Expired flows will not be queued for more than %d seconds",
	     readOnlyGlobals.sendTimeout);

  if((readOnlyGlobals.engineType != 0) || (readOnlyGlobals.engineId != 0))
    traceEvent(TRACE_INFO,
	       "Exported flows with engineType %d and engineId %d",
	       readOnlyGlobals.engineType, readOnlyGlobals.engineId);

  if(readOnlyGlobals.minFlowSize != 0)
    traceEvent(TRACE_INFO,
	       "TCP flows shorter than %u bytes will not be emitted",
	       readOnlyGlobals.minFlowSize);

  if(readOnlyGlobals.ignoreVlan)
    traceEvent(TRACE_INFO, "Vlan Ids will be ignored and set to 0.");

  if(readOnlyGlobals.ignoreProtocol)
    traceEvent(TRACE_INFO, "IP Protocol will be ignored and set to 0.");

  if(unlikely(readOnlyGlobals.ignoreIP))
    traceEvent(TRACE_INFO, "IP addresses will be ignored and set to 0.");

  if(readOnlyGlobals.ignorePorts)
    traceEvent(TRACE_INFO, "UDP/TCP src/dst ports will be ignored and set to 0.");

  if(readOnlyGlobals.ignoreTos)
    traceEvent(TRACE_INFO, "TCP TOS will be ignored and set to 0.");

#ifdef HAVE_GEOIP
  if(readOnlyGlobals.geo_ip_asn_db == NULL)
#endif
    traceEvent(TRACE_NORMAL, "Flows ASs will not be computed "
#ifndef HAVE_GEOIP
	       "(missing GeoIP support)"
#endif
);

  if((readOnlyGlobals.packetFlowGroup > 1) && (readOnlyGlobals.flowExportDelay == 0)) {
    traceEvent(TRACE_INFO, "WARNING: -B (%u) requires that you also set -e. Ignored.",
	       readOnlyGlobals.packetFlowGroup);
    readWriteGlobals->packetSentCount = 0;
  }

  if((readOnlyGlobals.packetFlowGroup > 0) && (readOnlyGlobals.flowExportDelay > 0))
    traceEvent(TRACE_INFO, "After %d flow packets are sent, we'll delay at least %d ms",
	       readOnlyGlobals.packetFlowGroup, readOnlyGlobals.flowExportDelay);
  else if(readOnlyGlobals.flowExportDelay > 0)
    traceEvent(TRACE_INFO, "The minimum intra-flow delay is of at least %d ms",
	       readOnlyGlobals.flowExportDelay);

  if(readOnlyGlobals.flowLockFile != NULL)
    traceEvent(TRACE_INFO,
	       "No flows will be sent if the lock file '%s' is present",
	       readOnlyGlobals.flowLockFile);

  if(readOnlyGlobals.numCollectors > 1) {
    if(readOnlyGlobals.reflectorMode)
      traceEvent(TRACE_INFO, "All flows will be sent to all defined "
		 "collectors (NetFlow reflector mode)");
    else
      traceEvent(TRACE_INFO, "Flows will be sent to the defined collectors "
		 "in round robin.");
  }

  traceEvent(TRACE_INFO, "Flows will be emitted in %s format",
	     readOnlyGlobals.netFlowVersion == 5 ? "NetFlow 5" :
	     (readOnlyGlobals.netFlowVersion == 9 ? "NetFlow 9" : "IPFIX"));

  if(readOnlyGlobals.pktSampleRate > 1)
    traceEvent(TRACE_INFO, "%sSampling packets at 1:%d rate",
	       readOnlyGlobals.fakePktSampling ? "Fake " : "",
	       readOnlyGlobals.pktSampleRate);

  if(readOnlyGlobals.flowSampleRate > 1) {
    traceEvent(TRACE_INFO, "Sampling flows at 1:%d rate", readOnlyGlobals.flowSampleRate);
    readWriteGlobals->flowsToGo = readOnlyGlobals.flowSampleRate;
  }

  if(readOnlyGlobals.use_vlanId_as_ifId != vlan_disabled) {
    traceEvent(TRACE_INFO, "Using %s VLAN Id as NetFlow interface Id",
	       (readOnlyGlobals.use_vlanId_as_ifId == inner_vlan) ? "inner" : "outer");
  } else {
    if(readOnlyGlobals.inputInterfaceIndex == NO_INTERFACE_INDEX)
      traceEvent(TRACE_INFO, "Flow input interface index is dynamic (last two MAC address bytes)");
    else
      traceEvent(TRACE_INFO, "Flow input interface index is set to %d",
		 readOnlyGlobals.inputInterfaceIndex);

    if(readOnlyGlobals.outputInterfaceIndex == NO_INTERFACE_INDEX)
      traceEvent(TRACE_INFO, "Flow output interface index is dynamic (last two MAC address bytes)");
    else
      traceEvent(TRACE_INFO, "Flow output interface index is set to %d",
		 readOnlyGlobals.outputInterfaceIndex);
  }

  if((readOnlyGlobals.pcapFile == NULL)
#ifdef HAVE_PF_RING
     && (readOnlyGlobals.metadata_fd == NULL)
#endif
     && (readOnlyGlobals.captureDev != NULL)) {
    if((readOnlyGlobals.pcapPtr == NULL)
#ifdef HAVE_PF_RING
       && (readWriteGlobals->ring == NULL)
#endif
       )
      traceEvent(TRACE_NORMAL, "Not capturing packet from interface (collector mode)");
    else
      traceEvent(TRACE_NORMAL, "Capturing packets from interface %s",
		 readOnlyGlobals.captureDev);
  }

  if(readOnlyGlobals.smart_udp_frags_mode)
    traceEvent(TRACE_NORMAL, "Smart fragment rebuild enabled (no fragments are rebuilt)");

  if(readOnlyGlobals.tunnel_mode)
    traceEvent(TRACE_NORMAL, "Enabled tunnel decoding (e.g. IPSEC/GTP)");

  readOnlyGlobals.nprobe_up = 1;
  dumpLogEvent(probe_started, severity_info, "nProbe started");

#ifndef WIN32
  if(readOnlyGlobals.becomeDaemon)
    daemonize();
#endif

  if((readOnlyGlobals.pcapFile == NULL) && (!readOnlyGlobals.nfLitePluginEnabled)
#ifdef HAVE_PF_RING
     && (readOnlyGlobals.metadata_fd == NULL)
#endif
     ) {
    dropPrivileges();
  }

  /* Change user-id then save the pid path */
#ifndef WIN32
  readOnlyGlobals.nprobePid = getpid();

  if(readOnlyGlobals.pidPath) {
    FILE *fd = fopen(readOnlyGlobals.pidPath, "w");
    if(fd != NULL) {
      fprintf(fd, "%lu\n", readOnlyGlobals.nprobePid);
      fclose(fd);
    } else
      traceEvent(TRACE_ERROR, "Unable to store PID in file %s",
		 readOnlyGlobals.pidPath);
  }
#endif

  load_mappings();

  if(readOnlyGlobals.quick_mode) {
    readOnlyGlobals.enableExtBucket = 0;
  } else {
    enablePlugins();
    setupPlugins();
  }

  if(readOnlyGlobals.enable_l7_protocol_discovery)
    initL7Discovery();

  if((readOnlyGlobals.pcapPtr
#ifdef HAVE_PF_RING
      || (readWriteGlobals->ring != NULL)
      || (readOnlyGlobals.metadata_fd != NULL)
#endif
      || (readOnlyGlobals.collectorInPort > 0)
      || readOnlyGlobals.nfLitePluginEnabled
      || readOnlyGlobals.logPluginEnabled
      || readOnlyGlobals.tracePerformance
      )) {
#if !defined(WIN32)
    pthread_attr_t tattr;
    struct sched_param param;

    /* initialized with default attributes */
    if(pthread_attr_init(&tattr) == 0) {
      /* safe to get existing scheduling param */
      if(pthread_attr_getschedparam (&tattr, &param) == 0) {
	param.sched_priority++; /* Increase priority */

	/* setting the new scheduling param */
	pthread_attr_setschedparam (&tattr, &param);
      }
    }
#endif

    /* Start a pool of threads */
    if((readOnlyGlobals.packetProcessThread = (pthread_t*)malloc(sizeof(pthread_t)*readOnlyGlobals.numProcessThreads)) == NULL) {
      traceEvent(TRACE_ERROR, "Not enough memory?");
      return(0);
    }

    if(readOnlyGlobals.collectorInPort > 0)
      createNetFlowListener(readOnlyGlobals.collectorInPort);

    if(unlikely(readOnlyGlobals.enable_debug)) {
      traceEvent(TRACE_WARNING, "*****************************************");
      traceEvent(TRACE_WARNING, "** You're running nprobe in DEBUG mode **");
      traceEvent(TRACE_WARNING, "*****************************************");
    }

    pthread_create(&readWriteGlobals->walkHashThread, NULL, hashWalker, NULL);
    traceEvent(TRACE_INFO, "Starting %u packet fetch thread(s)", readOnlyGlobals.numProcessThreads);
    pthread_create(&readWriteGlobals->dequeueThread, NULL, dequeueBucketToExport, NULL);

    if(readOnlyGlobals.pcapPtr
#ifdef HAVE_PF_RING
       || readWriteGlobals->ring
       || readOnlyGlobals.metadata_fd
#endif
       || readOnlyGlobals.tracePerformance
       ) {
      readWriteGlobals->numTerminatedFetchPackets = 0;

      if(readOnlyGlobals.pcapFileList != NULL) {
	struct fileList *fl = readOnlyGlobals.pcapFileList, *next;

	while(fl != NULL) {
	  if((openDevice(ebuf, 1, fl->path) == -1) || (readOnlyGlobals.pcapPtr == NULL))
	    traceEvent(TRACE_ERROR, "Unable to open file %s (%s)\n", fl->path, ebuf);
	  else {
	    if(readOnlyGlobals.pcapPtr) fetchPackets(NULL);
	  }

	  next = fl->next;
	  free(fl->path);
	  free(fl);
	  fl = next;
	}
      } else {
	pthread_start_routine fetcher;

#ifdef HAVE_PF_RING
	fetcher = (readWriteGlobals->ring || readOnlyGlobals.metadata_fd) ? fetchPfRingPackets : fetchPackets;
#else
	fetcher = fetchPackets;
#endif

	if(readOnlyGlobals.pcapFile != NULL) {
	  fetchPackets(NULL);
	} else {
	  /* Spawn fetcher thread */
	  u_long thread_id = 0;

	  if((readOnlyGlobals.numProcessThreads > 1) && readOnlyGlobals.quick_mode) {
	    traceEvent(TRACE_WARNING, "Quick mode can be used only in single threaded mode");
	    readOnlyGlobals.quick_mode = 0;
	  }

	  /* 1 receive thread */
	  pthread_create(&readOnlyGlobals.packetProcessThread[i],
#if !defined(WIN32)
			 &tattr,
#else
			 NULL,
#endif
			 fetcher, (void*)thread_id);

	  /* n process threads */
	  if(readOnlyGlobals.numProcessThreads > 1) {
	    for(i=0; i<readOnlyGlobals.numProcessThreads; i++) {
	      thread_id = i;

	      pthread_create(&readOnlyGlobals.packetProcessThread[i],
#if !defined(WIN32)
			     &tattr,
#else
			     NULL,
#endif
			     processPackets, (void*)thread_id);
	    }
	  }
	}
      }
    }
  }

  if(readOnlyGlobals.pcapFile) {
    u_int32_t i, tot_pkts = 0, tot_bytes = 0;

    for(i=0; i<readOnlyGlobals.numProcessThreads; i++)
      tot_pkts += readWriteGlobals->accumulateStats[i].pkts;

    traceEvent(TRACE_INFO, "No more packets to read. Sleeping...\n");

    // while(1) ntop_sleep(999); /* Sleep forever */
  } else {
    while(1) {
      // sleep(30); break;
      ntop_sleep(9999); /* Sleep forever */
    }
  }

  shutdown_nprobe();

  return(0);
}

/* ******************************** */
